#ifndef _MTDF_TCC
#define _MTDF_TCC
#include "osl/search/mtdf.h"
#include "osl/search/searchMoveVector.h"
#include "osl/search/shouldPromoteCut.h"
#include "osl/category/categoryListUtil.h"
#include "osl/move_classifier/pawnDropCheckmate.h"
#include "osl/move_classifier/moveAdaptor.h"
#include "osl/record/csa.h"
#ifdef OSL_SMP
#  include "osl/search/parallelSearch.h"
#  include "osl/misc/nonBlockDelete.h"
#endif
#include <stdexcept>
#include <iostream>

/* ------------------------------------------------------------------------- */
/*	Constructors							     */
/* ------------------------------------------------------------------------- */

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
MTDF(const HashEffectState& s, checkmate_t& c, Table& t, Recorder& r) 
  : base_t(r, &t), checkmate_searcher(&c),
    memory_tester(s, *checkmate_searcher, &t, r),
    timer(60)
{
  setInitialGuess();
  checkmate_searcher->setVerbose(base_t::table->isVerbose());
#ifdef OSL_SMP
  parallelSearch.setVerbose(t.isVerbose());
  memory_tester.shareCheckmateHolder(checkmate_holder);
#endif
}
template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
~MTDF() 
{
#ifdef OSL_SMP
  if (checkmate_holder.use_count() == 1)
    NonBlockDelete::reset(checkmate_holder);
#endif
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
void osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
setInitialGuess(int guess) 
{
  threshold = (guess & (~1)); 
  threshold -= eval::delta(memory_tester.state().getTurn());
  assert(threshold % 2);
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
void osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
setInitialGuess() 
{ 
  const Eval eval(memory_tester.state());
  setInitialGuess(eval.value());
}

#ifdef OSL_SMP
template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
void osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
setCheckmateSearcher(checkmate_t * c, const Worker *w)
{
  checkmate_searcher=c;
  memory_tester.setCheckmateSearcher(c, w);
}
#endif

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
bool osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
isReasonableMove(Move move, int pawn_sacrifice)
{
  return memory_tester.isReasonableMove(move, pawn_sacrifice);
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
const osl::misc::AlarmSwitch 
osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
alarmSwitch()
{ 
  return memory_tester.alarmSwitch();
}
template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
const osl::search::BigramKillerMove& 
osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
bigramKillerMove()
{
  return memory_tester.bigramKillerMove();
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
void osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
setBigramKillerMove(const BigramKillerMove& killers)
{
  memory_tester.setBigramKillerMove(killers);
}

/* ------------------------------------------------------------------------- */
/*	Methods for the toplevel loop					     */
/* ------------------------------------------------------------------------- */

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
const osl::Move osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
computeBestMove(int limit, const RealTime& shedule)
{
  const bool use_signal_timer = ! shedule.isInvalid();
  if (use_signal_timer)
  {
    timer = shedule;
    memory_tester.setTimeOut(timer.getTimeLeftInSeconds());
  }
  else
  {
    timer = RealTime(6000); // INFTY;
  }
  Move best_move;
#if OSL_SMP
  parallelSearch.setMasterAlarm(alarmSwitch());
#endif
  try
  {    
    if (memory_tester.state().getTurn()==BLACK)
      best_move = computeBestMoveOfPlayer<BLACK>(limit);
    else
      best_move = computeBestMoveOfPlayer<WHITE>(limit);
  }
  catch (...)
  {
#ifdef OSL_SMP
    parallelSearch.waitAll();
#endif
    if (use_signal_timer)
      memory_tester.resetTimeOut();
    throw;
  }
#ifdef OSL_SMP
  parallelSearch.waitAll();
#endif
  if (use_signal_timer)
    memory_tester.resetTimeOut();
  return best_move;
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
const osl::Move osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
computeBestMoveIteratively(int limit, const int step, 
			   const int initial_limit, int node_limit, 
			   const RealTime& initial_stop_shedule)
{
  setStopSchedule(initial_stop_shedule);
  double total_consumed = 0;
  double last_iteration_consumed = 0;

  base_t::recorder.resetNodeCount();
  int limit_iterative=initial_limit;
  Move best_move = Move::INVALID();
  try
  {
    RealTime current_timer = stopSchedule();
    best_move = computeBestMove(0, current_timer);
    while (limit_iterative < limit)
    {
      current_timer = stopSchedule();
      if (base_t::table->isVerbose())
	std::cerr << "iteration " << limit_iterative 
		  << " " << current_timer.getTimeLeftInDouble() << " left\n"
		  << std::flush;
      best_move = computeBestMove(limit_iterative, current_timer);
      if (base_t::table->isVerbose())
	std::cerr << "selected " << record::csa::show(best_move) << "\n";
      memory_tester.testTimeOut();
      if (best_move.isNormal())
      {
	bool time_over = false;
	last_iteration_consumed = timer.getConsumedInDouble();
	total_consumed += last_iteration_consumed;
	
	if (hasSchedule())
	{	
	  const double current_time_left
	    = stopSchedule().getTimeLeftInDouble();
	  if (current_time_left 
	      < last_iteration_consumed * base_t::next_iteration_coefficient)
	  {
	    time_over = true;
	    if (base_t::table->isVerbose())
	      std::cerr << "time limit " << current_time_left 
			<< " < " << last_iteration_consumed 
			<< " (*" << base_t::next_iteration_coefficient << ")\n";
	  }
	  if (! time_over)
	  {
	    SimpleHashRecord *record
	      = base_t::table->find(memory_tester.currentHash());
	    if (record)
	    {
	      record->setSumNodeCountOfChildren();
	      if (base_t::table->isVerbose())
		record->dumpNodeCount(std::cerr, 3);
	    }
	  }
	}
	// depth  200 䤹8ܤ餤ˤʤȤɤ
	// פ20ܤˤʤ뤳Ȥ⤢
	if (time_over || (base_t::recorder.nodeCount() *8 > node_limit))
	{
	  if (base_t::table->isVerbose())
	    std::cerr << "iteration stop at " << limit_iterative << "\n";
	  return best_move;
	}
      }
      limit_iterative += step;
    }
    if (base_t::table->isVerbose())
      std::cerr << "final iteration " << limit
		<< " " << stopSchedule().getTimeLeftInDouble() << " left\n";
    while (true)
    {
      best_move = computeBestMove(limit, stopSchedule());
      if (best_move.isNormal())
	break;
      memory_tester.testTimeOut();

      // ۤäƤλ
      if (limit >= 2000)
	break;

      limit += 200;
      if (base_t::table->isVerbose())
	std::cerr << "extend limit to " << limit << " before resign\n";
    }
  }
  catch (std::exception& e)
  {
    std::cerr << "std exception " << e.what() << "\n";
  }
  catch (...)
  {
    std::cerr << "unknown exception\n";
#ifndef NDEBUG
    throw;
#endif
  }
  if (base_t::table->isVerbose())
    std::cerr << "selected " << record::csa::show(best_move) 
	      << " at limit " << limit_iterative << "\n";
  return best_move;
}

/**
 * ᥤõ.
 * ǽ low/high fail ȽꤷѴؿ˸Ǥ
 */
template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
const osl::Move osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
computeBestMoveOfPlayer(int limit)
{
  if (base_t::isWinValue(P, threshold))
  {
    threshold = base_t::windowMax(P) - eval::delta(P);
  }
  assert(threshold % 2);
  base_t::recorder.startSearch(limit);
  if (base_t::table->isVerbose())
    std::cerr << "threshold " << threshold << "\n";

  SimpleHashRecord *record_in_table
    = base_t::table->allocate(memory_tester.currentHash(), limit);
  SimpleHashRecord *record = record_in_table;
  boost::scoped_ptr<SimpleHashRecord> record_if_not_allocated;
  if (! record_in_table)
  {
    record_if_not_allocated.reset(new SimpleHashRecord());
    record = record_if_not_allocated.get();
  }
  assert(record);
  assert(memory_tester.rootRecord() == 0
	 || memory_tester.rootRecord() == record);
  memory_tester.setRootRecord(record);

  // ޤơ֥򸣤
  SearchMove best_move_in_table;
  if (record_in_table)
  {
    best_move_in_table = record->bestMove();
    if (best_move_in_table.getMove().isNormal())
    {
      // ;Ǿ
      int table_value;
      if (record->template hasGreaterLowerBound<P>
	  (limit, base_t::winThreshold(P), table_value))
	return best_move_in_table.getMove();
    }
    if ((! best_move_in_table.getMove().isNormal())
	|| (best_move_in_table.record == 0))
    {
      best_move_in_table = SearchMove(); // invalidate
    }
  }

  // ƤΥƥμ򽸤
  SearchMoveSet& move_set = record->moves();
  while (true)
  {
    const category::CategoryEnv env(&memory_tester.state(), limit, 
				    &memory_tester.history(), 
				    memory_tester.getEval().progress32(),
				    record, 0,
				    &memory_tester.historyTable());
    CategoryListUtil::gatherAllMoves<MoveGenerator>(env, move_set);
    if (move_set.size() > 1 || limit > 2000 || limit == 0)
      break;
    limit += 100;
  }
  if (move_set.size() == 1)
    return move_set.front().getMove();

  const bool my_king_in_check
    = memory_tester.state().hasEffectBy(alt(P),memory_tester.state().template getKingPosition<P>());
#ifndef DONT_USE_CHECKMATE
  // ;ƤǤߤ root ǤƤǤʤ
  if (record_in_table)
  {
    const int check_limit 
      = ((limit < 500) 
	 ? 800 
	 : ((limit < 700) ? 3800 : 4000));
    const int checkmate_max = checkmate::limitToCheckCount(check_limit);
    const int checkmate_node = record->qrecord.checkmateNodesLeft(checkmate_max);
    if (checkmate_node > 0)
    {
      PathEncoding path = memory_tester.path();
      if (my_king_in_check)
      {
	// ꤫鲦꤬äƤ
	base_t::recorder.gotoCheckmateSearch(memory_tester.state(), checkmate_node);
	const bool lose = memory_tester.template isLosingState<P>(checkmate_node);
	base_t::recorder.backFromCheckmateSearch();
	if (lose)
	{
	  base_t::recordLoseByCheckmate(P, record);
	  return Move::INVALID();
	}
      }
      // ͤޤʤСͤޤȤͤ 
      {
	Move checkmate_move;
	base_t::recorder.gotoCheckmateSearch(memory_tester.state(), checkmate_node);
	const bool win = memory_tester.template isWinningState<P>
	  (checkmate_node, checkmate_move, record->qrecord.attack_oracle);
	base_t::recorder.backFromCheckmateSearch();
	if (win)
	{
	  base_t::recordWinByCheckmate(P, record, checkmate_move);
	  return checkmate_move;
	}
      }
      // ͤͤ
      if ((! my_king_in_check)
	  && (! record->threatmate().isThreatmate(P)))
      {
	Move threatmate_move;

	base_t::recorder.gotoCheckmateSearch(memory_tester.state(), checkmate_node);
	const bool threatmate 
	  = memory_tester.template isThreatmateState<P>
	  (checkmate_node, threatmate_move, record->qrecord.threatmate_oracle);
	base_t::recorder.backFromCheckmateSearch();
	if (threatmate)
	{
	  record->threatmate().setThreatmate(P, threatmate_move);
	  if (base_t::table->isVerbose())
	    std::cerr << "root threatmate " << threatmate_move << "\n";
	}
      }
    }
  }
#endif
  /**
   * threshold ۤȱۤʤä
   */
  SearchMoveSorter<P> lower_moves, higher_moves;

  /** 
   * ߤκɾ threshold ˤä upperbound ä
   * lower bound äꤹΤ threshold ѹ 
   */
  int current_value = base_t::minusInfty(P);
  upper_bound = base_t::minusInfty(alt(P));
  lower_bound = base_t::minusInfty(P);
  // ǽκ
  base_t::recorder.newCategory("table",limit);
  if (base_t::validTableMove(memory_tester.state(), best_move_in_table.moveLogProb(), limit))
  {
    SearchMove trying_move = best_move_in_table;
    trying_move.setLogProbAtMost(Probabilities::TableMove);
    int val = base_t::minusInfty(P);
    for (int i=0; i<2; ++i)	// 2ޤbestmove¤γ
    {
      if (i > 0)
      {
	if (base_t::isWinValue(alt(P), val))
	  break;
	threshold = EvalTraits<P>::max(current_value - eval::delta(P)
				       + Eval::captureValue(newPtypeO(P,PAWN)),
				       base_t::winThreshold(alt(P)));
	current_value = base_t::minusInfty(P);
      }
      if (testMove<P>(trying_move, val, current_value, limit))
      {
	if (base_t::isWinValue(alt(P), current_value))
	  return trying_move.getMove();
	threshold = current_value + eval::delta(P);
	// record->bestMoveϥɥ쥹Ȥʤ
	SearchMove *stored_move = move_set.find(trying_move.getMove());
	higher_moves.insert(std::make_pair(current_value, stored_move));
	break;
      }
    }
    if (higher_moves.empty() && (! base_t::isWinValue(alt(P), val)))
    {
      threshold = current_value - eval::delta(P);
      current_value = base_t::minusInfty(P);
    }
  }
  // 礤κϿ
  const int qvalue
    = memory_tester.template quiescenceValue<P>(base_t::winThreshold(alt(P)),
						base_t::winThreshold(P));
  if (limit == 0)
    return record->qrecord.bestMove();
  
  if (higher_moves.empty())
  {
    const int safe_threshold 
      = qvalue + Eval::captureValue(newPtypeO(P,PAWN))*2 - eval::delta(P);
    if (EvalTraits<P>::betterThan(threshold, safe_threshold))
    {
      threshold = safe_threshold;
      current_value = base_t::minusInfty(P);
      threshold = EvalTraits<P>::max(threshold, current_value+eval::delta(P));
    }
  }
  if (record->qrecord.bestMove().isNormal())
  {
    const MoveLogProb quiescence_move(record->qrecord.bestMove(),
				      Probabilities::KillerMove-1);
    move_set.assignIfBetter(quiescence_move);
  }
  // ٤Ƥμ
  SearchMoveVector all_moves;
  {
    SearchMoveSet::range r(move_set);
    for (SearchMoveSet::iterator p=r.first; p!=r.last; ++p)
    {
      if (move_classifier::MoveAdaptor<move_classifier::PawnDropCheckmate<P> >
	  ::isMember(memory_tester.state(), p->getMove()))
	continue; // 
      if (! my_king_in_check)
	if (p->getMove().isPass()
	    || ShouldPromoteCut::canIgnoreAndNotDrop(p->getMove()))
	  continue;

      if (higher_moves.empty() 	// ʣʤ
	  || (p->getMove() != higher_moves.begin()->second->getMove()))
      {
	p->setLogProbAtMost(1000);	// ϳ̵褦
	if (p->getMove().isNormal() && (p->getLogProb() <= limit))
	  all_moves.push_back(&*p);
      }
    }
  }
  all_moves.sortByValue<P>(threshold + Eval::captureValue(newPtypeO(P,PAWN))*2);
  Move best_move;
  examineMoves<P>("uniqued moves", current_value, limit, all_moves,
		  lower_moves, higher_moves);
  if (higher_moves.empty()) // low fail
  {
    best_move 
      = retryByLowerF<P>(record_in_table, limit, lower_moves, higher_moves);
  } 
  else 
  {
    // high fail: loop ƬǤ threshold Ͱʾμʣǧ
    best_move = selectBestMoveByHigherF<P>(record_in_table, limit, 
					   lower_moves, higher_moves);
  }
#ifdef OSL_SMP
  base_t::recorder.setCheckmateCount(checkmate_holder->totalNodeCount());
#else
  base_t::recorder.setCheckmateCount(checkmate_searcher->totalNodeCount());
#endif
  const double consumed = timer.getConsumedInDouble();
  base_t::recorder.finishSearch(best_move, consumed,base_t::table->isVerbose());
  return best_move;
}

#ifdef OSL_SMP
namespace osl
{
  namespace search
  {
    struct MtdfParShared
    {
      const SearchMoveVector moves;
      vector<int> values;
      vector<int> thresholds;
      int current_value;
      int threshold;
      int lower_bound;
      /** true if not all moves are examined */
      bool stopped;

      LightMutex mutex;

      MtdfParShared(const SearchMoveVector& m, int initial_values,
		    int current_value, int threshold, int lb)
	: moves(m), values(moves.size(), initial_values),
	  thresholds(moves.size(), 0), current_value(current_value), 
	  threshold(threshold), lower_bound(lb), stopped(false)
      {
      }
    };

    template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities,typename framework_t,Player Turn>
    class MtdfParHelper : public JobContent
    {
      framework_t f; // copy of framework
      typedef typename framework_t::checkmate_t checkmate_t;
      typedef CheckmateSearcherHolder<checkmate_t> holder_t;
      boost::shared_ptr<MtdfParShared> shared;
      holder_t *checkmate_holder;
      int limit;
      int index, last;
    public:
      MtdfParHelper(int priority,
		    const boost::shared_ptr<MtdfParShared>& shared,
		    framework_t const& f1, holder_t *holder,
		    int limit, int index, int last)
	: JobContent(priority), f(f1), shared(shared),
	  checkmate_holder(holder),
	  limit(limit), index(index), last(last) {
      }
      void setIndex(int newIndex) { index=newIndex;}
      void operator()(void) {
	checkmate_t *my_checkmate = checkmate_holder->clone(getWorker());
	f.setCheckmateSearcher(my_checkmate, getWorker());
	f.template examineMovesParBlock<Turn>(shared, limit, index, last);
	setFinished(true);
      }
    };
  }
}


template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
void osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
examineMovesParBlock(const boost::shared_ptr<MtdfParShared>& shared,
		     int limit, size_t index, size_t last)
{
  assert(last <= shared->moves.size());
  assert(last - index > 0);
  int my_current_value;
  {
    LightMutex::scoped_lock lk(shared->mutex);
    threshold= shared->threshold;
    lower_bound = shared->lower_bound;
    my_current_value= shared->current_value;
  }
  if (base_t::isWinValue(P,my_current_value)) {
    shared->stopped = true;
    return;
  }
  const int relax_value = Eval::captureValue(newPtypeO(alt(P),PAWN))/2/8;
  if (last - index == 1) {
    int val;
    shared->thresholds[index]=threshold;
    if (EvalTraits<P>::betterThan(threshold, upper_bound))
      upper_bound = (threshold + relax_value)&(~1);
    if (testMove<P>(*(shared->moves[index]), val, my_current_value, limit))
    {
      threshold = my_current_value + eval::delta(P);
      {
	LightMutex::scoped_lock lk(shared->mutex);
	if(EvalTraits<P>::betterThan(threshold, shared->threshold))
	  shared->threshold=threshold;
	if(EvalTraits<P>::betterThan(my_current_value, shared->current_value))
	  shared->current_value=my_current_value;
	if(EvalTraits<P>::betterThan(lower_bound, shared->lower_bound))
	  shared->lower_bound=lower_bound;
      }
    }
    shared->values[index]=val;
    return;
  }
  typedef MtdfParHelper<Eval,MoveGenerator,Table,Recorder,Probabilities,MTDF,P> helper_t;
  int priority=index;
  boost::shared_ptr<helper_t> 
    helper(new helper_t(priority, shared, *this, checkmate_holder.get(), 
			limit, index+1, last));
  Job job(helper);
  Worker *worker=parallelSearch.getWorker();
  {
    Job current = worker->getCurrentJob();
    current.setChildJob(helper);
  }
  while (last - index > 0) {
    int my_current_value;
    {
      LightMutex::scoped_lock lk(shared->mutex);
      my_current_value= shared->current_value;
      threshold= shared->threshold;
      lower_bound = shared->lower_bound;
    }
    if (base_t::isWinValue(P, my_current_value)){
      job.stopNow();
      shared->stopped = true;
      return;
    }
    if (last - index > 1)
      worker->push_back(job);
    int val;
    shared->thresholds[index]=threshold;
    bool ret=false;
    try {
      if (EvalTraits<P>::betterThan(threshold, upper_bound))
	upper_bound = (threshold + relax_value)&(~1);
      ret=testMove<P>(*(shared->moves[index]), val, my_current_value, limit);
    }
    catch (StopNow& e){
      shared->stopped = true;
      std::cerr << "break worker " << worker->threadId() << "\n";
      if (!worker->pop_back_job()) { // stolen
	job.stopNow();
      }
      throw e;
    }
    if (ret)
    {
      threshold = my_current_value + eval::delta(P);
      {
	LightMutex::scoped_lock lk(shared->mutex);
	if(EvalTraits<P>::betterThan(threshold, shared->threshold))
	  shared->threshold=threshold;
	if(EvalTraits<P>::betterThan(my_current_value, shared->current_value))
	  shared->current_value=my_current_value;
	if(EvalTraits<P>::betterThan(lower_bound, shared->lower_bound))
	  shared->lower_bound=lower_bound;
      }
    }
    shared->values[index]=val;
    ++index;
    if (last - index > 0) {
      if (worker->pop_back_job()) { // not stolen
	helper->setIndex(index+1);
	helper->setFinished(false);
	helper->setPriority(index);
      }
      else{ // wait for the result
#ifdef ONE_THREAD_PER_PROC
	DecActiveHelper dec_lock(parallelSearch);
	while (!helper->isFinished()) {
	  Job another_job;
	  try
	  {
	    another_job=parallelSearch.getJobNonBlock(worker);
	  }
	  catch (StopNow&)
	  {
	    helper->stopNow();
	    throw;
	  }
	  if (! another_job.isNull()) {
	    try{
	      SetCurrentJobLock(*worker,another_job);
	      IncActiveHelper inc_lock(parallelSearch);
	      another_job();
	    }
	    catch (StopNow& e) {
	      std::cerr << "ignore StopNow  "
			<< &another_job << "\n";
	    }
	  }
	  else {
	    boost::thread::yield();
	  }
	}
#else
	helper->waitResult();
#endif
	return;
      }
    }
  }
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
void osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
examineMovesPar(int& current_value, int limit, const SearchMoveVector& moves, 
		SearchMoveSorter<P>& lower_moves, SearchMoveSorter<P>& higher_moves) 
{
  boost::shared_ptr<MtdfParShared> shared
    (new MtdfParShared(moves, base_t::winThreshold(alt(P)), current_value,
		       threshold, lower_bound));
  const int block_size = 8;
  for (size_t first=0; first<moves.size(); first+=block_size)
  {
    if (base_t::isWinValue(P, current_value))
    {
      assert(! higher_moves.empty());
      return;
    }
    const size_t last = std::min(moves.size(), first+block_size);
    examineMovesParBlock<P>(shared, limit, first, last);
    current_value= shared->current_value;
    threshold= shared->threshold;
    lower_bound = shared->lower_bound;
    bool all_even = true;
    for (size_t i=first; i<last; i++) {
      if (EvalTraits<P>::betterThan(shared->values[i],shared->thresholds[i]))
	higher_moves.insert(std::make_pair(shared->values[i], moves[i]));
      if (! shared->thresholds[i] % 2)
	all_even = false;
    }
    memory_tester.testTimeOut();
    assert(shared->stopped || all_even);
    if (higher_moves.empty())
    {
      for (size_t i=first; i<last; i++)
	lower_moves.insert(std::make_pair(shared->values[i], moves[i]));
      typename SearchMoveSorter<P>::const_reverse_iterator best = lower_moves.rbegin();
#ifndef __APPLE__
      assert(best != lower_moves.rend());
#endif
      threshold = best->first - eval::delta(P);
      threshold += Eval::captureValue(newPtypeO(P,PAWN)); // 2礵˲
      threshold = EvalTraits<P>::max(threshold, base_t::winThreshold(alt(P)));
      current_value = threshold - eval::delta(P);
    }
  }
}

#endif

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
void osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
examineMoves(const char *category_name, 
	     int& current_value, int limit, const SearchMoveVector& moves, 
	     SearchMoveSorter<P>& lower_moves, SearchMoveSorter<P>& higher_moves) 
{
  base_t::recorder.newCategory(category_name, limit);
#ifdef OSL_SMP
  if (moves.size()>1)
  {
    examineMovesPar<P>(current_value, limit, moves, lower_moves, higher_moves);
    return;
  }
#endif
  for (SearchMoveVector::const_iterator p=moves.begin(); p!=moves.end(); ++p)
  {
    if (base_t::isWinValue(P, current_value))
    {
      assert(! higher_moves.empty());
      return;
    }
    int val;
    if (testMove<P>(**p, val, current_value, limit))
    {
      threshold = current_value + eval::delta(P);
      higher_moves.insert(std::make_pair(current_value, *p));
      if ((moves.size() > 30) && base_t::table->isVerbose())
      {
	std::cerr << ".";
      }
    }
    if (higher_moves.empty() 
	&& (! base_t::isWinValue(alt(P), val)))
    {
      // low fail ʤΤǤƤ
      lower_moves.insert(std::make_pair(val, *p));
      threshold = val - eval::delta(P); // ɤɤ f 򲼤Ƹǻ
      threshold += Eval::captureValue(newPtypeO(P,PAWN)); // 2礵˲
      threshold = EvalTraits<P>::max(threshold, base_t::winThreshold(alt(P)));
      current_value = threshold - eval::delta(P);
    }
  }
  assert(higher_moves.isValidAll());
}
  
/**
 * ȥåץ٥ 롼 ( low fail )
 * 
 * threshold 򲼤ʤõthreshold ꤬ۤǤ selectBestMoveByHigherF 
 */
template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
const osl::Move osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
retryByLowerF(SimpleHashRecord *record, int limit, 
	      SearchMoveSorter<P>& lower_moves, SearchMoveSorter<P>& higher_moves)
{
  assert(higher_moves.empty());
  assert(lower_moves.isValidAll());
  SearchMoveVector moves;
  while (higher_moves.empty() && (lower_moves.size() > 1))
  {
    base_t::recorder.recordTopLevelLowFail(lower_moves.begin()->second->moveLogProb(),
					   threshold);
    if (base_t::table->isVerbose())
      std::cerr << "low fail " << threshold << "\n";
    upper_bound = lower_moves.begin()->first;
    const int old_threshold = threshold;
    for (typename SearchMoveSorter<P>::const_iterator p=lower_moves.begin(); p!=lower_moves.end(); ++p)
      moves.push_back(p->second);
    lower_moves.clear();
    // 餱λؤϿʤΤ lower moves Ϥ󸺤äƤ椯ǽ
    int current_value = base_t::minusInfty(P);
    examineMoves<P>("low fail", current_value, limit, 
		    moves, lower_moves, higher_moves);
    moves.clear();
    threshold += Eval::captureValue(newPtypeO(P,PAWN))*4; // 8礵˲
    threshold = EvalTraits<P>::max(threshold, base_t::winThreshold(alt(P)));
    if (old_threshold == threshold)
      return Move::INVALID();
  }

  if (! higher_moves.empty())
    return selectBestMoveByHigherF<P>(record, limit, 
				      lower_moves, higher_moves);

  if (lower_moves.size() == 0)
    return Move::INVALID();

  assert(lower_moves.size() == 1);
  const SearchMove *best_move = lower_moves.begin()->second;
  const int val = lower_moves.begin()->first;
  assert(limit >= 0);
  if (record)
    base_t::recordLowerBound(P, record, limit, *best_move, val);

  // thresholdϤΤޤ: 򿶤櫓Τ¸
  return best_move->getMove();
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
const osl::Move osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
selectBestMoveByHigherF(SimpleHashRecord *record,
			int limit, SearchMoveSorter<P>& lower_moves, 
			SearchMoveSorter<P>& higher_moves)
{
  assert(higher_moves.isValidAll());
  assert(lower_moves.isValidAll());
  base_t::recorder.recordTopLevelHighFail(higher_moves.begin()->second->moveLogProb(),
					  threshold);
  // lower_moves ǽΤ򽦤
  {
    assert(! higher_moves.empty());
    int current_value = higher_moves.begin()->first;
    threshold = current_value + eval::delta(P);
    SearchMoveVector moves_maybe_beyond_f;
    for (typename SearchMoveSorter<P>::const_iterator p=lower_moves.begin(); 
	 p!=lower_moves.end(); ++p)
    {
      // p->first ˤ ¤äƤ롤 lower_moves Ͼ¤ι⤤¤Ǥ
      if (! EvalTraits<P>::betterThan(p->first, threshold))
	break;
      moves_maybe_beyond_f.push_back(p->second);
    }
    examineMoves<P>("high fail", current_value, limit,
		    moves_maybe_beyond_f, lower_moves, higher_moves);
  }
  const int triage_value = Eval::captureValue(newPtypeO(alt(P),PAWN))/2;
  if (higher_moves.size() > 32)
    triageMovesSpeculatively<P>(higher_moves, limit, triage_value*8);
  if (higher_moves.size() > 16)
    triageMovesSpeculatively<P>(higher_moves, limit, triage_value*4);
  triageMovesSpeculatively<P>(higher_moves, limit, triage_value);
  triageMovesSpeculatively<P>(higher_moves, limit, triage_value/8);

  while (higher_moves.size() > 1)
  {
    if (base_t::table->isVerbose())
    {
      std::cerr << "higher " << higher_moves.size() << ":\t";
      higher_moves.summary(4);
    }
    typename SearchMoveSorter<P>::iterator p = higher_moves.begin();
    int best_value = p->first;
    if (base_t::isWinValue(P, best_value))
    {
      // ʬ⤱Ф⤦ɤμǤɤ(ߤϵͤߤ򸫤ĤȤ)
      break;
    }

    SearchMoveSorter<P> next_jobs;
    next_jobs.insert(*p);
    while (++p != higher_moves.end())
    {
      threshold = best_value + eval::delta(P);
      if (EvalTraits<P>::betterThan(threshold, upper_bound))
      {
	// relax upper_bound (in case for search instability)
	upper_bound = (threshold + triage_value/8)&(~1);
      }

      const SearchMove *target = p->second;
      SearchMove move_to_try = *target;
      if (higher_moves.size() <= 3)
	move_to_try.setLogProbAtMost(Probabilities::ReSearch);

      int val,max_value=threshold-eval::delta(P);
      if (testMove<P>(move_to_try, val, max_value, limit))
      {
	best_value = val;
	next_jobs.insert(std::make_pair(val, target)); // ȥå׸
	if (base_t::isWinValue(P, best_value))
	{
	  // ʬ⤱Ф⤦ɤμǤɤ(ߤϵͤߤ򸫤ĤȤ)
	  break;
	}
      }
    }
    higher_moves.swap(next_jobs);
  }
  const SearchMove *best_move = higher_moves.begin()->second;
  const int best_value = higher_moves.begin()->first;
  assert(limit >= 0);
  if (record)
  {
    base_t::recordLowerBound(P, record, limit, *best_move, best_value);
    assert((record->bestMove().getMove() == best_move->getMove())
	   || (record->lowerLimit() == limit && record->lowerBound() == best_value)
	   || (limit <= std::max(record->lowerLimit(), record->upperLimit())));
  }

  // thresholdϤΤޤ: 򿶤櫓Τ¸
  return best_move->getMove();
}

/**
 * ͤͿ higher_moves ʤ.
 * ޤ̤٤Τ绨Ĥͤ򤢤Ƥ
 * upperbound ޤäƤ鿿ܤ
 */
template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
void osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
triageMovesSpeculatively(SearchMoveSorter<P>& higher_moves, int limit,
			 int jump)
{
  while (higher_moves.size() > 1)
  {
    SearchMoveSorter<P> next_jobs;
    typename SearchMoveSorter<P>::iterator p = higher_moves.begin();
    const int best_value = p->first;
    threshold = best_value + jump + eval::delta(P);
    if (base_t::isWinValue(P, threshold)
	|| EvalTraits<P>::betterThan(threshold, upper_bound))
      return;

    if (base_t::table->isVerbose())
    {
      std::cerr << "triage " << higher_moves.size() << " " << threshold << ": ";
      higher_moves.summary(4);
    }
    next_jobs.insert(*p);	// ǽμϻˤ
    while (++p != higher_moves.end())
    {
      const SearchMove *target = p->second;
      SearchMove move_to_try = *target;
      if (higher_moves.size() <= 3)
	move_to_try.setLogProbAtMost(Probabilities::ReSearch);

      int val;
      int max_value=threshold-eval::delta(P);
      if (testMove<P>(move_to_try, val, max_value, limit))
      {
	next_jobs.insert(std::make_pair(val, target)); // ͹
	next_jobs.insert(++p, higher_moves.end());
	break;
      }
    }
    if (next_jobs.size() <= 1)
      return;
    
    higher_moves.swap(next_jobs);
  }
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
bool osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
testMove(const SearchMove& move, int& val, int& max_value, int limit)
{
  assert(P == move.getMove().player());
  assert(P == memory_tester.state().getTurn());
  assert(EvalTraits<P>::betterThan(threshold,max_value));
  assert(threshold % 2);
  assert(! base_t::isWinValue(BLACK, threshold));
  assert(! base_t::isWinValue(WHITE, threshold));
  assert(move.getMove().isNormal());

  if ((! memory_tester.timerAvailable()) && (! timer.isInvalid()) && (! timer.timeLeft()))
    throw misc::NoMoreTime();
#ifdef OSL_SMP
  parallelSearch.checkStop();
#endif

  // adjusted_move: logp  history table 
  SearchMove adjusted_move = move;
  const int history_logp = memory_tester.historyTable().logp(move.getMove());
  if (history_logp < move.getLogProb())
    adjusted_move = SearchMove(MoveLogProb(move.getMove(), history_logp),
			       move.record);
  lower_bound = EvalTraits<P>::min(lower_bound, 
				   Eval::captureValue(newPtypeO(alt(P),ROOK)));
  upper_bound = EvalTraits<P>::max(upper_bound, 
				   Eval::captureValue(newPtypeO(P,ROOK)));
  assert(EvalTraits<P>::betterThan(threshold, lower_bound));
  assert(EvalTraits<P>::betterThan(upper_bound, threshold));
  memory_tester.setRootBounds(P, lower_bound, upper_bound);
  val = memory_tester.template nullWindowSearch<P>(threshold, adjusted_move,
						   limit);
  move.record = adjusted_move.record;
  
  if (EvalTraits<P>::betterThan(val,max_value))
  {
    max_value = val;
    if (EvalTraits<P>::betterThan(val,threshold))
    {
      lower_bound = EvalTraits<P>::max(val, lower_bound);
      memory_tester.setKillerMove(move.getMove());
      return true;
    }
    // fall thru
  }
  return false;
}

#endif /* _MTDF_TCC */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
