#ifndef _PARALLEL_SEARCH_H
#define _PARALLEL_SEARCH_H
#include "osl/misc/lightMutex.h"
#include "osl/misc/atomicCounter.h"
#include "osl/misc/alarm.h"
#include "osl/stl/vector.h"
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <boost/thread.hpp>
#include <deque>
#include <stdexcept>

#ifndef NOT_ONE_THREAD_PER_PROC
/** ĤΥץå˺륹åɤϤ1Ĥˤ */
#  define ONE_THREAD_PER_PROC
#endif

namespace osl
{
  namespace search
  {
    typedef LightMutex Mutex;

    class Worker;
    class Job;
    struct StopNow : std::runtime_error
    {
      StopNow();
    };
    class JobContent
    {
    public:
      static const int DefaultPriority=100;
      static const int NotFoundPriority=100000;
    private:
      boost::condition condition;
      boost::mutex mutex;
      Worker* worker;
      volatile bool finished;
      int priority;
      volatile bool must_stop_now;
      boost::weak_ptr<JobContent> child;
    public:
      JobContent() : worker(0), finished(0), priority(DefaultPriority),
		     must_stop_now(false)
      {
      }
      explicit JobContent(int priority)
	: worker(0), finished(0), priority(priority),
	  must_stop_now(false)
      {
	assert(priority<NotFoundPriority);
      }
      virtual ~JobContent();
      virtual void operator()(void)=0;
      void setFinished(bool val) {
	boost::mutex::scoped_lock lk(mutex);
	finished=val;
	condition.notify_all();
      }
      bool isFinished() const{
	return finished;
      }
      void setWorker(Worker* w) { worker=w; }
      const Worker *getWorker() const { return worker; }
      void checkStop();
      void waitResult();
      int getPriority() const{
	return priority;
      }
      void setPriority(int p) {
	priority=p;
      }
      void stopNow();
      void setChildJob(const boost::shared_ptr<JobContent>& new_child) {
	child = new_child;
      }
      bool stopping() const {
	return must_stop_now;
      }
    };
    class ExitNullJobLock
    {
      boost::shared_ptr<JobContent> job_content;
    public:
      ExitNullJobLock(const boost::shared_ptr<JobContent>& job_content)
	: job_content(job_content) {}
      ~ExitNullJobLock() {
	job_content->setWorker(0);
	job_content->setFinished(true);
      }
    };
    class Job
    {
      boost::shared_ptr<JobContent> job_content;
    public:
      Job() {}
      explicit Job(const boost::shared_ptr<JobContent>& job_content) 
	: job_content(job_content) {
      }
      bool isNull() const{
	return !job_content;
      }
      void operator()(void) {
	ExitNullJobLock lock(job_content);
	assert(!isNull());
	try{
	  (*job_content)();
	}
	catch (StopNow& e) {
	}
      }
      void setWorker(Worker* w) {
	job_content->setWorker(w);
      }
      void setFinished(bool val) {
	job_content->setFinished(val);
      }
      void waitResult() {
	job_content->waitResult();
      }
      int getPriority() const{
	return job_content->getPriority();
      }
      void stopNow() const {
	job_content->stopNow();
      }
      void checkStop() const {
	job_content->checkStop();
      }
      void setChildJob(const boost::shared_ptr<JobContent>& new_child) {
	if (job_content)
	  job_content->setChildJob(new_child);
      }
    };
    class JobPool
    {
      std::deque<Job> jobs;
    public:
      void push_back(Job job) {
	jobs.push_back(job);
      }
      bool pop_back() {
	if (jobs.empty())
	  return false;
	jobs.pop_back();
	return true;
      }
      int getPriority() const{
	if (jobs.empty()) 
	  return JobContent::NotFoundPriority;
	return jobs.front().getPriority();
      }
      /** call with shared_mutex locked */
      const Job getJob(Worker *worker) {
	if (jobs.empty()) 
	  return Job();
	Job ret=jobs.front();
	jobs.pop_front();
	ret.setWorker(worker);
	return ret;
      }
      void clear(){
	jobs.clear();
      }
    };
    class ParallelSearch;
    /**
     * 
     */
    class Worker
    {
      ParallelSearch* parent;
      boost::thread *th;
      boost::shared_ptr<JobPool> job_pool;
      Job current_job;
      unsigned int thread_id;
    public:
      explicit Worker(ParallelSearch* parent,boost::thread* th=0) 
	: parent(parent),th(th),job_pool(new JobPool()),
	  thread_id(static_cast<unsigned int>(-1))
      {
      }
      void operator()(void);
      /** call with shared_mutex locked */
      const Job getJob(Worker *w) {
	Job ret=job_pool->getJob(w);
	return ret;
      }
      int getPriority() const{
	return job_pool->getPriority();
      }
      boost::thread* getThread() const{
	return th;
      }
      /** shared_mutex  lock  */
      void push_back(Job job);
      /** shared_mutex  lock  */
      bool pop_back_job();
    public:      
      void checkStop() const;
      const Job getCurrentJob() const{
	return current_job;
      }
      void setCurrentJob(Job const& job) {
	current_job=job;
      }
      void resetCrurrentJob() {
	current_job=Job();
      }
      unsigned int threadId() const { return thread_id; }
      friend class ParallelSearch;

      void stopMaster();
    private:
      void setThreadId(unsigned int id) { thread_id = id; }
      /** shared_mutex  lock ֤ǸƤ */
      void clearJobs() { job_pool->clear(); };
    };

    class SetCurrentJobLock
    {
      Worker& worker;
      Job backup;
    public:
      SetCurrentJobLock(Worker& w,Job const& job) :worker(w),backup(w.getCurrentJob()){
	worker.setCurrentJob(job);
      }
      ~SetCurrentJobLock(){
	worker.setCurrentJob(backup);
      }
    };

    struct DecActiveHelper{
      ParallelSearch & parallel_search;
      explicit DecActiveHelper(ParallelSearch & parallel_search);
      ~DecActiveHelper();
    };
    struct IncActiveHelper{
      ParallelSearch & parallel_search;
      explicit IncActiveHelper(ParallelSearch & parallel_search);
      ~IncActiveHelper();
    };
    struct IncWaitingHelper{
      ParallelSearch & parallel_search;
      explicit IncWaitingHelper(ParallelSearch & parallel_search);
      ~IncWaitingHelper();
    };
    struct DecWaitingHelper{
      ParallelSearch & parallel_search;
      explicit DecWaitingHelper(ParallelSearch & parallel_search);
      ~DecWaitingHelper();
    };

    class ParallelSearch
    {
      AtomicCounter num_actives;
      AtomicCounter max_num_actives;
      AtomicCounter thread_count;
      AtomicCounter waiting; // WorkerΥȥåץ٥ԤäƤ륹åɿ
      vector<Worker*> workers;
      int num_cpus;
      time_t last_time;
      AlarmSwitch master_stop_flag;
      bool verbose;
      volatile bool stop_all;
      void checkTime();
    public:
      boost::condition shared_condition;
      boost::mutex shared_mutex;

      ParallelSearch();
      ~ParallelSearch();

      void setVerbose(bool v=true) { verbose = v; }
      void addInLock(Worker* worker) {
	workers.push_back(worker);
	worker->setThreadId(workers.size()-1);
      }
      void add(Worker* worker) {
	boost::mutex::scoped_lock lk(shared_mutex);
	addInLock(worker);
      }
      const Job getJob(Worker *w);
      const Job getJobInLock(Worker *w);
      const Job getJobNonBlock(Worker *w);
      /**
       * ߤΥåɤWorker֤
       * бΤʤϡallocate롥
       */
      Worker* getWorker();
      void notify_one() {
	shared_condition.notify_one();
      }
      void makeWorkers();
      void setNumCPUs(int n) {
	num_cpus=n;
      }
      int numCpus() const { return num_cpus; }
      void checkStop();
      bool stopping() const { 
	return stop_all 
	  || (master_stop_flag.notify
	      && *master_stop_flag.notify == Alarm::TIMEOUT);
      }
      void waitAll();

      void setMasterAlarm(AlarmSwitch alarm) { master_stop_flag = alarm; }
      bool stopMaster();
    private:
      friend class JobContent;
      friend class IncActiveHelper;
      friend class DecActiveHelper;
      friend class IncWaitingHelper;
      friend class DecWaitingHelper;
      void incActives();
      void decActives();
      void incWaiting();
      void decWaiting();
    };

    extern osl::search::ParallelSearch parallelSearch;
  } // namespace search
} // namespace osl
#endif /* _PARALLEL_SEARCH_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
