/* SparseRegressionMultiplier.h
 */
#ifndef _STAT_SPARSEREGRESSIONMULTIPLIER_H
#define _STAT_SPARSEREGRESSIONMULTIPLIER_H

#include "osl/stat/matrixMultiplier.h"

namespace osl
{
  namespace stat
  {
    struct DoubleReader
    {
      virtual ~DoubleReader() {}
      virtual double read() = 0;
    };
    struct DoubleConstReader : public DoubleReader
    {
      const double value;
      explicit DoubleConstReader(double v) : value(v)
      {
      }
      double read()
      {
	return value;
      }
    };
    struct DoubleReferenceReader : public DoubleReader
    {
      const double& value;
      explicit DoubleReferenceReader(double& v) : value(v)
      {
      }
      double read()
      {
	return value;
      }
    };
    struct DoubleArrayReader : public DoubleReader
    {
      const double *array;
      unsigned int cur;
      explicit DoubleArrayReader(const double *a) : array(a), cur(0)
      {
      }
      double read()
      {
	return array[cur++];
      }
    };
    template <class Iter>
    struct DoubleIteratorReader : public DoubleReader
    {
      Iter it;
      explicit DoubleIteratorReader(Iter i) : it(i)
      {
      }
      double read()
      {
	return *it++;
      }
    };
    

    /**
     * incrementaly compute out=Ap
     * where A = X^tX
     */
    class SparseRegressionMultiplier : public MatrixMultiplier
    {
      /** ĴŤߤ dimension.  οľܴʤ */
      unsigned int m_dimA;
      /** regularization term */
      double m_lambda;
    public:
      // CREATORS
      explicit SparseRegressionMultiplier(unsigned int dimA=0, double lambda=0.0);
      ~SparseRegressionMultiplier();
      // MANIPULATORS
    protected:
      void init(unsigned int new_dimA) { m_dimA = new_dimA; }
    public:
      // ACCESSORS
      void prod(const double *p, double *out) const;
      unsigned int dim() const;
      /**
       * @return false if the end of sequence (the next call must return
       *   the beginning of the sequence)
       */
      virtual bool getVectorX(unsigned int& num_elements,
			      unsigned int *non_zero_indices, 
			      double *non_zero_values) const =0;
      virtual void newIteration() const;
    protected:
      /** partially compute X^tX*p and add to out where non_zero_{indices,values} reperesent one row in X */
      void addProduct(unsigned int non_zeros,
		      const unsigned int *non_zero_indices, 
		      const double *non_zero_values,
		      const double *p, double *out) const;
    public:
      /**
       * @param xty is b such that Ax=b to solve x
       * @param diag_inv used in order to make diagonal preconditioner of X^tX
       * @return #instances
       */
      unsigned int computeXtY(DoubleReader& y, double *xty, double *diag_inv) const;
    };    

    /**
     * feature ͤ 1.0 ˸ꤵƤС
     */
    class BinarySparseRegressionMultiplier : public MatrixMultiplier
    {
      unsigned int m_dimA;
      double m_lambda;
    public:
      // CREATORS
      explicit BinarySparseRegressionMultiplier(unsigned int dimA,
						double lambda=0.0);
      ~BinarySparseRegressionMultiplier();
      // MANIPULATORS
      // ACCESSORS
      void prod(const double *p, double *out) const;
      unsigned int dim() const;
      /**
       * @return false if the end of sequence (the next call must return
       *   the beginning of the sequence)
       */
      virtual bool getVectorX(unsigned int& num_elements,
			      unsigned int *non_zero_indices) const =0;
      virtual void newIteration() const;
      /** partially compute X^tX*p and add to out where non_zero_indices reperesent one row in X */
      void addProduct(unsigned int non_zeros,
		      const unsigned int *non_zero_indices, 
		      const double *p, double *out) const;
    };    
    
  } // namespace stat
} // namespace osl


#endif /* _STAT_SPARSEREGRESSIONMULTIPLIER_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
