/* csaClient.cc
 */
#include "osl/game_playing/csaClient.h"
#include "osl/game_playing/gnuShogiClient.h"
#include "osl/game_playing/gameState.h"
#include "osl/game_playing/csaLogger.h"
#include "osl/game_playing/csaStopwatch.h"
#include "osl/game_playing/moveWithComment.h"
#include "osl/record/csa.h"
#include "osl/record/csaIOError.h"
#include "osl/sennichite.h"

#include <iostream>

osl::game_playing::
CsaClient::CsaClient(ComputerPlayer *black, ComputerPlayer *white,
		     CsaLogger *l, std::istream& is, std::ostream& os)
  : CuiClient(black, white, l, is, os),
    show_move_with_comment(false), silent(false)
{
  setComputerPlayer(WHITE, true);
}

osl::game_playing::
CsaClient::~CsaClient()
{
}

void osl::game_playing::
CsaClient::readAndProcessCommand()
{
  char ctime_buf[64];
  if (! silent) {
    std::cerr << "\nCsaClient start waiting ";
    const time_t now = time(0);
    std::cerr << ctime_r(&now, ctime_buf) << std::endl;
  }
  CsaStopwatch timer;
  std::string line;
  std::getline(is, line);
  if (! silent) {
    std::cerr << "\nCsaClient read " << line << " ";
    const time_t now = time(0);
    std::cerr << ctime_r(&now, ctime_buf) << std::endl;
  }
  const long op_think_time = timer.read();
  if (! is)
  {
    const char *message = "istream error (maybe closed)";
    std::cerr << message << std::cerr;
    logger->writeComment(message);
    throw EndGame();
  }
  
  if (line == "%TORYO")
  {
    logger->resign(state->state().getTurn());
    throw EndGame();
  }

  // TODO: %MATTA, %CHUDAN
  try
  {
    const Move op_move=record::csa::strToMove(line, state->state());
    const GameState::MoveType illegal_move = state->isIllegal(op_move);
    if (illegal_move)
    {
      std::cerr << "illegal move: " << line << "\n";
      logger->inputError(line.c_str());
      if (illegal_move == GameState::PAWN_DROP_FOUL)
	logger->writeComment("pawn drop foul");
      else if (illegal_move == GameState::UNSAFE_KING)
	logger->writeComment("unsafe king");
      else if (illegal_move == GameState::OTHER_INVALID)
	logger->writeComment("other illegal move");
      throw EndGame();
    }
    const Sennichite result = pushMove(MoveWithComment(op_move), op_think_time);
    if (! result.isNormal())
    {
      os << "%SENNICHITE" << std::endl;
      logger->endByRepetition(result);
      throw EndGame();
    }
  }
  catch (record::csa::CsaIOError&)
  {
    std::cerr << "bad input: " << line << "\n";
    logger->inputError(line.c_str());
    throw EndGame();
  }
  return;
}

void osl::game_playing::
CsaClient::setShowMoveWithComment(bool value)
{
  show_move_with_comment = value;
}

void osl::game_playing::
CsaClient::processComputerMove(const MoveWithComment& selected, 
			       int my_think_time)
{
  static std::string reserved="+7776FU";
  const Move best_move = selected.move;
  if ((! best_move.isNormal())
      || (state->isIllegal(best_move)))
  {
    if (best_move == Move::DeclareWin())
    {
      os << "%KACHI\n";
      logger->endByDeclaration(state->state().getTurn());
    }
    else
    {
      os << "%TORYO\n";
      logger->resign(state->state().getTurn());
    }
    throw EndGame();
  }

  os << record::csa::show(best_move, reserved);
  if (show_move_with_comment)
  {
    os << ",'* " << selected.value;
    for (vector<Move>::const_iterator p=selected.moves.begin(); 
	 p!=selected.moves.end(); ++p)
    {
      os << " ";
      os << record::csa::show(*p, reserved);
    }    
  }
  os << std::endl << std::flush;

  assert(isComputer(state->state().getTurn()));

  const Sennichite result = pushMove(selected, my_think_time);
  if (! result.isNormal())
  {
    logger->endByRepetition(result);
    throw EndGame();
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
