/* speculativeAllMoves.cc
 */
#include "osl/game_playing/speculativeAllMoves.h"
#include "osl/game_playing/gameState.h"
#include "osl/game_playing/searchPlayer.h"
#include "osl/search/timeControl.h"
#include "osl/search/searchRecorder.h"
#include "osl/search/shouldPromoteCut.h"
#include "osl/search/searchMoveVector.h"
#include "osl/search/simpleHashTable.h"
#include "osl/category/categoryListUtil.h"
#include "osl/category/standardMoveGenerator.h"
#include "osl/misc/realTime.h"
#include "osl/record/csa.h"
#include "osl/sennichite.h"
#include "osl/misc/lightMutex.h"
#include "osl/misc/nonBlockDelete.h"
#include <boost/thread.hpp>
#include <exception>
#include <iostream>
#include <ctime>
#include <unistd.h>

osl::game_playing::SpeculativeAllMoves::
ResultVector::ResultVector()
{
}

osl::game_playing::SpeculativeAllMoves::
ResultVector::~ResultVector()
{
}

void osl::game_playing::SpeculativeAllMoves::ResultVector::
add(Move prediction, const MoveWithComment& result)
{
  Mutex::scoped_lock lk(mutex);
  data.push_back(std::make_pair(prediction, result));
}
const osl::game_playing::MoveWithComment* osl::game_playing::SpeculativeAllMoves::ResultVector::
find(Move prediction) const
{
  Mutex::scoped_lock lk(mutex);
  for (vector_t::const_iterator p=data.begin(); p!=data.end(); ++p) {
    if (p->first == prediction)
      return &p->second;
  }
  return 0;
}
void osl::game_playing::SpeculativeAllMoves::ResultVector::
clear()
{
  Mutex::scoped_lock lk(mutex);
  data.clear();
}

/* ------------------------------------------------------------------------- */

struct osl::game_playing::SpeculativeAllMoves::SearchAllMoves::StatusLock
{
  volatile Status& status;
  const Status in, out;
  StatusLock(volatile Status *s, Status i, Status o) : status(*s), in(i), out(o)
  {
    status = in;
  }
  StatusLock(volatile Status *s, Status i) : status(*s), in(i), out(*s)
  {
    status = in;
  }
  ~StatusLock()
  {
    status = out;
  }
};

struct osl::game_playing::SpeculativeAllMoves::SearchAllMoves::Generator
{
  GameState& state;
  SearchPlayer& player;
  MoveVector moves;
  volatile Status& status;
  int index;
  Generator(GameState& s, SearchPlayer& p, volatile Status& st) 
    : state(s), player(p), status(st), index(-1)
  {
  }
  static int convert(Player player, int value)
  {
    if (player == BLACK)
      return value;
    return -value;
  }
  typedef std::multimap<int, Move, std::greater<int> > map_t;
  static void insert(map_t& sorter, int value, Move move)
  {
    sorter.insert(std::make_pair(value, move));
  }
  void pickUpFromRecord(const HashKey& key, const SimpleHashRecord& record,
			const SimpleHashTable& table)
  {
    map_t sorter;
    const Player turn = key.turn();
    MoveVector qmoves;
    record.qrecord.loadMoves(qmoves);
    for (size_t i = 0; i < qmoves.size(); i++)
    {
      const HashKey new_key = key.newHashWithMove(qmoves[i]);
      const SimpleHashRecord *new_record = table.find(new_key);
      if (! new_record)
	continue;
      if (new_record->lowerLimit() >= 0)
	insert(sorter, convert(turn, new_record->lowerBound()), qmoves[i]);
      else if (new_record->upperLimit() >= 0)
	insert(sorter, convert(turn, new_record->upperBound()), qmoves[i]);
    }
    SearchMoveSet::const_range r(record.moves());
    for (SearchMoveSet::const_iterator p=r.first; p!=r.last; ++p)
    {
      if (! p->record || ! p->record->hasLowerBound(0))
	continue;
      
      insert(sorter, convert(turn, p->record->lowerBound()), p->getMove());
    }
    map_t::const_iterator last = sorter.end();
    if (record.qrecord.hasStaticValue())
    {
      const int base_value = convert(turn, record.qrecord.staticValue());
      last = sorter.lower_bound(base_value);
    }
    for (map_t::const_iterator p=sorter.begin(); p != last; ++p)
    {
      if (std::find(moves.begin(), moves.end(), p->second) != moves.end())
	continue;
      moves.push_back(p->second);
    }
  }
  void pickUpMoves()
  {
    try
    {
      MoveWithComment result;
      {
	StatusLock lk(&status, PREDICTION2);
	result = player.selectBestMove(state, 1, 0);
      }
      if (result.move.isNormal())
      {
	moves.push_back(result.move);
      }
      const search::SimpleHashTable* table = player.table();
      if (! table)
	return;
      const HashKey key = state.state().getHash();
      const SimpleHashRecord *record = table->find(key);
      if (! record)
	return;
      pickUpFromRecord(key, *record, *table);
#ifndef NDEBUG
      if (moves.size())
      {
	std::cerr << "search prediction ";
	for (MoveVector::const_iterator p=moves.begin(); p!=moves.end(); ++p)
	  std::cerr << record::csa::show(*p);
	std::cerr << "\n";
      }
#endif
    }
    catch (std::exception& e)
    {
      std::cerr << "Generator::bestMove " << e.what() << "\n";
    }
  }
  void generateOtherMoves()
  {
    using namespace search;
    SearchMoveSet move_set;
    const category::CategoryEnv env(&state.state(), 1000, 
				    &state.moveHistory());
    CategoryListUtil::gatherAllMoves<StandardMoveGenerator>(env, move_set);
    SearchMoveVector all_moves;
    {
      const Player turn = state.state().getTurn();
      const bool my_king_in_check
	= state.state().hasEffectBy(alt(turn),
				    state.state().getKingPosition(turn));
      SearchMoveSet::range r(move_set);
      for (SearchMoveSet::iterator p=r.first; p!=r.last; ++p)
      {
	if (! my_king_in_check)
	  if (p->getMove().isPass()
	      || ShouldPromoteCut::canIgnoreAndNotDrop(p->getMove()))
	    continue;
	if (std::find(moves.begin(), moves.end(), p->getMove()) != moves.end())
	  continue;
	all_moves.push_back(&*p);
      }
    }
    all_moves.sortByProbability();
    for (SearchMoveVector::const_iterator p=all_moves.begin(); p!=all_moves.end();
	 ++p)
    {
      moves.push_back((*p)->getMove());
    }
  }
  const Move nextMove()
  {
    if (index < 0)
    {
      pickUpMoves();
      generateOtherMoves();
    }
    ++index;
    if (index < (int)moves.size())
      return moves[index];
    std::cerr << "finish\n";
    return Move::INVALID();
  }
};

osl::game_playing::SpeculativeAllMoves::
SearchAllMoves::SearchAllMoves(ResultVector& r)
  : results(r), current_move(Move::INVALID()), status(INITIAL), seconds(-1),
    stop_flag(false)
{
}

osl::game_playing::SpeculativeAllMoves::
SearchAllMoves::~SearchAllMoves()
{
}

void osl::game_playing::SpeculativeAllMoves::
SearchAllMoves::setUp(const GameState& main_state,
		      const SearchPlayer& main_player,
		      int standard_seconds, bool has_byoyomi)
{
  NonBlockDelete::reset(player);
  try
  {
    player.reset(dynamic_cast<SearchPlayer*>(main_player.clone()));
    player->setVerbose(false);
    player->setNextIterationCoefficient(has_byoyomi ? 1 : 2); // default=4 Ⱦʬ
    state = main_state.clone();
    generator.reset(new Generator(*state, *player, status));
    seconds = standard_seconds;
  }
  catch (std::exception& e)
  {
    player.reset();
    std::cerr << "setUp " << e.what() << "\n";
    throw;
  }
}

void osl::game_playing::SpeculativeAllMoves::
SearchAllMoves::run()
{
  if (! player)
    return;
  
  StatusLock lk(&status, RUNNING, FINISHED);
  while (true) 
  {
    boost::thread::yield(); // test whether opponent's move has arrived
    if (stop_flag)
      return;
    Move prediction;
    {
      StatusLock lk(&status, PREDICTION1);
      prediction = generator->nextMove();
    }
    if (! prediction.isNormal())
      return;
    if (stop_flag)
      return;
    const MoveWithComment result = testMove(prediction);
    results.add(prediction, result);
  }
}

const osl::game_playing::MoveWithComment osl::game_playing::SpeculativeAllMoves::
SearchAllMoves::testMove(Move predicted_move)
{
  StatusLock lk(&status, SEARCH1);
  {
    Mutex::scoped_lock lk(mutex);
    current_move = predicted_move;
  }
  assert(state);
  state->pushMove(predicted_move);
  assert(player);
  player->pushMove(predicted_move);
  MoveWithComment result(Move::INVALID());
  std::cerr << "prediction (" << seconds << ") "
	    << record::csa::show(predicted_move) << " ";
  const time_t now = time(0);
  char ctime_buf[64];
  std::cerr << ctime_r(&now, ctime_buf);
  try
  {
    StatusLock lk(&status, SEARCH2);
    result = (seconds > 0)
      ? player->searchWithSecondsForThisMove(*state, seconds)
      : player->selectBestMove(*state, -1, -1); // infty
    if (! stop_flag)
      std::cerr << record::csa::show(result.move) << "\n";
  }
  catch (std::exception& e)
  {
    // TODO table full clear?
    std::cerr << "error in speculative thread " << e.what() << "\n";
    stop_flag = true;
    NonBlockDelete::deleteAll();
  }
  catch (...)
  {
    std::cerr << "error in speculative thread\n";
    stop_flag = true;
  }
  state->popMove();
  player->popMove();
  {
    Mutex::scoped_lock lk(mutex);
    current_move = Move::INVALID();
  }
  return result;
}

void osl::game_playing::SpeculativeAllMoves::
SearchAllMoves::stopOtherThan(Move the_move)
{
  stop_flag = true;
  if (currentMove() != the_move)
    stopNow();
  else
  {
    std::cerr << "match " << record::csa::show(the_move) << "\n";
    assert(player);
    player->setVerbose(true);
  }
}
      
void osl::game_playing::SpeculativeAllMoves::
SearchAllMoves::stopNow()
{
  stop_flag = true;
  waitRunning();
  if (player && status != FINISHED)
  {
    std::cerr << "stopNow " << status << "\n";
    const bool success
      = player->stopSearchNow();
    if (! success)
      std::cerr << "stop search failed\n";
  }
}

void osl::game_playing::SpeculativeAllMoves::
SearchAllMoves::waitRunning()
{
  while (status == INITIAL)
  {
    if (!player)
      return;
    timespec rq = { 0, 20000000 }, rm;
    while (nanosleep(&rq, &rm) == -1) 
    {
      rq = rm;
    }
  }
}

void osl::game_playing::SpeculativeAllMoves::
SearchAllMoves::setStopSchedule(const misc::RealTime& timer)
{
  if (player && status != FINISHED)
  {
    waitRunning();
    player->setStopSchedule(timer);
  }
}

const osl::Move osl::game_playing::SpeculativeAllMoves::
SearchAllMoves::currentMove() const
{
  Mutex::scoped_lock lk(mutex);
  return current_move;
}

/* ------------------------------------------------------------------------- */


struct osl::game_playing::SpeculativeAllMoves::Runner
{
  SpeculativeAllMoves *parent;
  Runner(SpeculativeAllMoves *p) : parent(p)
  {
  }
  void operator()()
  {
    parent->searcher->run();
  }
};

osl::game_playing::SpeculativeAllMoves::
SpeculativeAllMoves()
  : results(new ResultVector()), last_search_seconds(-1), has_byoyomi(false),
    allowed(true)
{
}

osl::game_playing::
SpeculativeAllMoves::~SpeculativeAllMoves()
{
  stopAll();
  selectBestMoveCleanUp();
}

void osl::game_playing::SpeculativeAllMoves::
startSpeculative(const boost::shared_ptr<GameState> state,
		 const SearchPlayer& main_player)
{
  if (! allowed)
    return;
  
  try
  {
    results->clear();
    searcher.reset(new SearchAllMoves(*results));
    searcher->setUp(*state, main_player, last_search_seconds, has_byoyomi);
    thread.reset(new boost::thread(Runner(this)));
  }
  catch (std::exception& e)
  {
    std::cerr << "startSpeculative " << e.what();
    searcher.reset();
  }
}

void osl::game_playing::SpeculativeAllMoves::
clearResource()
{
  assert(! thread);
  searcher.reset();
}

void osl::game_playing::SpeculativeAllMoves::
stopOtherThan(Move the_move)
{
  if (searcher)
    searcher->stopOtherThan(the_move);
}

void osl::game_playing::SpeculativeAllMoves::
stopAll()
{
  if (searcher)
    searcher->stopNow();
}

const osl::game_playing::MoveWithComment
osl::game_playing::SpeculativeAllMoves::
waitResult(Move last_move, int wait_for,
	   SearchPlayer& main_player, int byoyomi)
{
  if (! allowed || ! searcher)
    return MoveWithComment(Move::INVALID());

  last_search_seconds = wait_for;
  has_byoyomi = (byoyomi > 0);
  
  const time_t start_time = time(0);
  assert(last_move.isNormal());
  const MoveWithComment *result = 0;
  bool stop_now = false;
  if (searcher->currentMove() != last_move)
  {
    stop_now = true;
    wait_for = 0;
  }
  const time_t stop_time = start_time + wait_for;

  if (searcher->isFinished())
  {
    if ((result = results->find(last_move)))
      goto found;
    return MoveWithComment(Move::INVALID());
  }
  searcher->setStopSchedule(misc::RealTime(wait_for));
  char ctime_buf[64];
  std::cerr << "wait for (" << wait_for << ") "
	    << ctime_r(&stop_time, ctime_buf);

  while (true)
  {
    if ((result = results->find(last_move)))
      break;
    assert(searcher);
    if (searcher->isFinished())
      return MoveWithComment(Move::INVALID());
    
    if (! stop_now)
    {
      time_t now = time(0);
      stop_now = now >= stop_time;
    }
    if (stop_now) {
      searcher->stopNow();
    }
    boost::thread::yield();
    timespec rq = { 0, 100000000 }, rm;
    nanosleep(&rq, &rm);
  }
found:
  if (! searcher->isFinished())
    searcher->stopNow();
  if (result->move.isNormal()) {
    SearchPlayer *player = searcher->currentPlayer();
    if (player)
      main_player.swapTable(*player);
  }
  return *result;
}
  
void osl::game_playing::SpeculativeAllMoves::
selectBestMoveCleanUp()
{
  if (! thread)
    return;
    
  if (searcher && ! searcher->isFinished())
    searcher->stopNow();
  thread->join();
  thread.reset();
  if (searcher)
    NonBlockDelete::reset(searcher);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
