/* speculativeSearchPlayer.cc
 */
#include "osl/game_playing/speculativeSearchPlayer.h"
#include "osl/game_playing/speculativeAllMoves.h"
#include "osl/game_playing/gameState.h"
#include "osl/game_playing/searchPlayer.h"
#include "osl/container/moveStack.h"
#include "osl/sennichite.h"
#include "osl/misc/nonBlockDelete.h"
#include <iostream>
#include <ctime>
#include <unistd.h>


osl::game_playing::SpeculativeSearchPlayer::
SpeculativeSearchPlayer(Player my_turn, SearchPlayer *player)
  : main_player(player), 
    speculative(new SpeculativeAllMoves()),
    my_turn(my_turn)
{
}

osl::game_playing::
SpeculativeSearchPlayer::~SpeculativeSearchPlayer()
{
}

osl::game_playing::ComputerPlayer* 
osl::game_playing::SpeculativeSearchPlayer::clone() const
{
  return new SpeculativeSearchPlayer(my_turn, 
				     dynamic_cast<SearchPlayer*>(main_player->clone()));
}

void osl::game_playing::SpeculativeSearchPlayer::
setMaxThreads(int new_max_threads)
{
  speculative->setMaxThreads(new_max_threads);
}

void osl::game_playing::SpeculativeSearchPlayer::
pushMove(Move m)
{
  main_player->pushMove(m);
  if (m.player() == my_turn)
  {
    if (previous_state.get() && speculative_search_allowed)
    {
      try
      {
	previous_state->pushMove(m);
	speculative->startSpeculative(previous_state, *main_player);
      }
      catch (std::exception& e)
      {
	std::cerr << e.what() << " in SpeculativeSearchPlayer::pushMove\n";
	speculative->clearResource();
      }
      NonBlockDelete::reset(previous_state);
    }
  }
  else
  {
    if (speculative_search_allowed)
      speculative->stopOtherThan(m);
  }
}

void osl::game_playing::SpeculativeSearchPlayer::
popMove()
{
  main_player->popMove();
  previous_state.reset();
  speculative->stopAll();
}

bool osl::game_playing::SpeculativeSearchPlayer::
stopSearchNow()
{
  return main_player->stopSearchNow();
}

int osl::game_playing::SpeculativeSearchPlayer::
standardSearchSeconds(const GameState& state, int total_left, int byoyomi) const
{
  return 
    std::max(1, main_player->secondsForThisMove(state, total_left, byoyomi)-1);
}

const osl::game_playing::MoveWithComment
osl::game_playing::SpeculativeSearchPlayer::
selectBestMove(const GameState& state, int total_left, int byoyomi)
{
  const time_t start_time = time(0);
  MoveWithComment result = MoveWithComment(Move::INVALID());
  const Move last_move = state.moveHistory().lastMove();

  const int wait_for = standardSearchSeconds(state, total_left, byoyomi);
  
  if (last_move.isNormal())
    result = speculative->waitResult(last_move, wait_for, *main_player,
				     byoyomi);

  const time_t now = time(0);
  char ctime_buf[64];
  if (result.move.isNormal()) {
#ifdef DEBUG_SPECULATIVE_EXECUTION
    std::cerr << "returned " << record::csa::show(result.move) 
	      << " " << ctime_r(&now, ctime_buf);
#endif
    selectBestMoveCleanUp(state);
    return result;
  }
  std::cerr << "search again " << ctime_r(&now, ctime_buf);

  const int consumed = (now - start_time);
  if (total_left)
    total_left = std::max(1, total_left - consumed);
  if (byoyomi)
    byoyomi = std::max(1, byoyomi - consumed);

  result = main_player->selectBestMove(state, total_left, byoyomi);
  selectBestMoveCleanUp(state);
  return result;
}

void osl::game_playing::SpeculativeSearchPlayer::
selectBestMoveCleanUp(const GameState& state)
{
  try
  {
    previous_state = state.clone();
  }
  catch (std::exception& e)
  {
    std::cerr << e.what() << std::endl;
    previous_state.reset();
  }
  speculative->selectBestMoveCleanUp();
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
