#include "osl/record/opening/openingBook.h"
#include "osl/record/record.h"
#include "osl/record/compactBoard.h"
#include "osl/apply_move/applyMove.h"
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_int.hpp>
#include <algorithm>
#include <iostream>
#include <stdexcept>

osl::record::opening::
WinCountBook::WinCountBook(const char *filename)
  : ifs(filename)
{
  if (! ifs)
  {
    const char *message = "WinCountBook: open failed ";
    std::cerr << message << filename << std::endl;
    throw std::runtime_error(std::string(message) + filename);
  }
  nStates=readInt();
}

osl::record::opening::
WinCountBook::~WinCountBook()
{
}

int osl::record::opening::
WinCountBook::readInt()
{
  int ret=0;
  CArray<char,4> cs;
  ifs.read(&cs[0],4);
  for (int i=0;i<4;i++) {
    ret = (ret<<8)|(cs[i]&255);
  }
  return ret;
}

void osl::record::opening::
WinCountBook::seek(int offset)
{
  ifs.seekg(offset,std::ios::beg);
}

osl::vector<osl::record::opening::OBMove> osl::record::opening::
WinCountBook::getMoves(int stateIndex)
{
  assert(stateIndex >= 0);
  seek(4+16*stateIndex+8);
  int nMoves=readInt();
  int moveIndex=readInt();
  seek(4+16*nStates+8*moveIndex);
  vector<OBMove> moves;
  moves.reserve(nMoves);
  for(int i=0;i<nMoves;i++)
  {
    Move move=Move::makeDirect(readInt());
    int stateIndex=readInt();
    moves.push_back(OBMove(move,stateIndex));
  }
  return moves;
}

int osl::record::opening::
WinCountBook::getWinCount(int stateIndex)
{
  seek(4+16*stateIndex);
  return readInt();
}

int osl::record::opening::
WinCountBook::getLoseCount(int stateIndex)
{
  seek(4+16*stateIndex+4);
  return readInt();
}

std::ostream& osl::record::opening::operator<<(std::ostream& os, const WMove& w)
{
  osl::record::writeInt(os, osl::record::opening::OMove(w.getMove()));
  osl::record::writeInt(os, w.getStateIndex());
  osl::record::writeInt(os, w.getWeight());
  return os;
}

std::istream& osl::record::opening::operator>>(std::istream& is, WMove& w)
{
  w.move = OMove(osl::record::readInt(is)).operator Move();
  w.stateIndex = osl::record::readInt(is);
  w.weight = osl::record::readInt(is);
  return is;
}


osl::record::opening::
WeightedBook::WeightedBook(const char *filename)
  : ifs(filename)
{
  if (! ifs)
  {
    const char *message = "WeightedBook: open failed ";
    std::cerr << message << filename << std::endl;
    throw std::runtime_error(std::string(message) + filename);
  }
#ifndef NDEBUG
  int version = 
#endif
    readInt(ifs);
  assert(version == 1);
  nStates = osl::record::readInt(ifs);
  nMoves = osl::record::readInt(ifs);
  startState = osl::record::readInt(ifs);
}

osl::record::opening::
WeightedBook::~WeightedBook()
{
}

void osl::record::opening::
WeightedBook::seek(int offset)
{
  ifs.seekg(offset,std::ios::beg);
}

osl::vector<osl::record::opening::WMove> osl::record::opening::
WeightedBook::getMoves(int stateIndex)
{
  assert(stateIndex >= 0);
  seek(HEADER_SIZE + STATE_SIZE * stateIndex);
  int moveIndex=readInt(ifs);
  int nWMoves=readInt(ifs);
  seek(HEADER_SIZE + STATE_SIZE * nStates + MOVE_SIZE * moveIndex);
  vector<WMove> moves;
  moves.reserve(nWMoves);
  for(int i=0;i<nWMoves;i++)
  {
    WMove wm;
    ifs >> wm;
    moves.push_back(wm);
  }
  return moves;
}

osl::SimpleState osl::record::opening::
WeightedBook::getBoard(int stateIndex)
{
  seek(HEADER_SIZE + STATE_SIZE * nStates + MOVE_SIZE * nMoves
       + BOARD_SIZE * stateIndex);
  CompactBoard board;
  ifs >> board;
  return board.getState();
}

int osl::record::opening::
WeightedBook::getWhiteWinCount(int stateIndex)
{
  seek(HEADER_SIZE + STATE_SIZE * stateIndex);
  readInt(ifs);
  readInt(ifs);
  return readInt(ifs);
}

int osl::record::opening::
WeightedBook::getBlackWinCount(int stateIndex)
{
  seek(HEADER_SIZE + STATE_SIZE * stateIndex);
  readInt(ifs);
  readInt(ifs);
  readInt(ifs);
  return readInt(ifs);
}

void osl::record::opening::
WeightedBook::validate()
{
  {
    osl::SimpleState state(osl::HIRATE);
    osl::SimpleState start = getBoard(startState);
    assert(state == start);
  }

  bool visited[nStates];
  for (int i = 0; i < nStates; i++)
  {
    visited[i] = false;
  }

  osl::vector<int> stateToCheck;
  stateToCheck.push_back(startState);
  visited[startState] = true;

  while (!stateToCheck.empty())
  {
    const int index = stateToCheck.back();
    stateToCheck.pop_back();
    osl::SimpleState state = getBoard(index);
    osl::vector<osl::record::opening::WMove> moves = getMoves(index);
    for (size_t i = 0; i < moves.size(); i++)
    {
      osl::SimpleState newState(state);
      osl::ApplyMoveOfTurn::doMove(newState, moves[i].getMove());
      const int nextIndex = moves[i].getStateIndex();

      osl::SimpleState stateInFile = getBoard(nextIndex);
      assert(newState == stateInFile);
      if (!visited[nextIndex])
      {
	stateToCheck.push_back(nextIndex);
	visited[nextIndex] = true;
      }
    }
  }
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
