/* tableAnalyzer.cc
 */
#include "osl/search/analyzer/tableAnalyzer.h"
#include "osl/search/analyzer/recordSet.h"
#include "osl/search/simpleHashTable.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/search/searchMoveList.h"
#include "osl/eval/evalCompareLarger.h"
#include "osl/record/csa.h"

#include <map>
#include <iostream>
#include <sstream>
#include <iomanip>

namespace osl
{
  namespace search
  {
    namespace analyzer
    {
      const int INVALID_VALUE = -1; // FIXME:  Ȥꤢ
      struct MoveEntry 
      {
	Move best_move;
	int upper_bound, lower_bound;
	int upper_limit, lower_limit;
	const SimpleHashRecord *record;
	MoveEntry(Move b, const SimpleHashRecord *r) 
	  : best_move(b), 
	    upper_bound(INVALID_VALUE), lower_bound(INVALID_VALUE),
	    upper_limit(0), lower_limit(0),
	    record(r)
	{
	  if (record)
	  {
	    upper_limit = record->upperLimit();
	    if (upper_limit > 0)
	      upper_bound = record->upperBound();
	    lower_limit = record->lowerLimit();
	    if (lower_limit > 0)
	      lower_bound = record->lowerBound();
	  }
	}
      };
      struct MoveMap 
	: public std::multimap<int,MoveEntry,eval::EvalCompareLargerNT> 
      {
	typedef std::multimap<int,MoveEntry,eval::EvalCompareLargerNT> map_t;
	explicit MoveMap(Player player) : map_t(player)
	{
	}
      };

      struct TopLevelAnalyzer
      {
	const SimpleHashTable *table;
	MoveMap sorter;
	int num_print_moves;
	std::ostream& os;
	Player player;
	TopLevelAnalyzer(Player p, 
			 const SimpleHashTable *t, int num, std::ostream& o) 
	  : table(t), sorter(p), num_print_moves(num), os(o), player(p)
	{
	}
	void examineAll(const SimpleHashRecord *record);
	void showMoves();
      };

      const std::string limitToString(int limit)
      {
	if (limit < 0)
	  return "*";
	std::ostringstream ss;
	ss << limit;
	return ss.str();
      }
      const std::string valueToString(int value)
      {
	if (value == INVALID_VALUE)
	  return "*";
	std::ostringstream ss;
	ss << value;
	return ss.str();
      }

    } // namespace analyzer
  } // namespace search
} // namespace osl

void osl::search::analyzer::
TopLevelAnalyzer::showMoves()
{
  int num_printed = 0;
  for (MoveMap::const_iterator p=sorter.begin(); 
       (p!=sorter.end())&&(num_printed < num_print_moves); ++p, ++num_printed)
  {
    os 
#if 0
      << std::setw(5) << p->first << " "
#endif
      << "[" 
      << std::setw(5) << valueToString(p->second.lower_bound) << "(" 
      << std::setw(4) << limitToString(p->second.lower_limit) << ") < " 
      << std::setw(5) << valueToString(p->second.upper_bound) << "("
      << std::setw(4) << limitToString(p->second.upper_limit) << ") " 
      << "] \t";
    csaShow(os, p->second.best_move);
    os << " ";
    SearchMoveList best_moves;
    TableAnalyzer::makeBestMoveList(p->second.record, best_moves, 16);
    for (SearchMoveList::const_iterator p=best_moves.begin(); 
	 p!=best_moves.end(); ++p)
    {
      if (! p->getMove().isPass())
      {
	csaShow(os, p->getMove());
	os << " "; // << p->getLogProb() << " ";
      }
      else if (p->getLogProb() == SearchTable::CheckmateSpecialDepth)
      {
	os << "CHECKMATE";
      }
      else
      {
	// 礤
	os << "*";
      }
    }
    os << "\n";
  }
}

void osl::search::analyzer::
TopLevelAnalyzer::examineAll(const SimpleHashRecord *record)
{
  assert(record);
  const SearchMoveSet& moves = record->moves();
  {
    SearchMoveSet::const_range r(moves);
    for (SearchMoveSet::const_iterator p=r.first; p!=r.last; ++p)
    {
      const SearchMove& cur_move = *p;
      /** ɽϿƤɤȽ˻ȤäƤ */
      int value=INVALID_VALUE;
      const SimpleHashRecord *next_record = cur_move.record;
      if (cur_move.getMove() == record->bestMove().getMove())
      {
	if (next_record && next_record->hasUpperBound(1))
	  value = next_record->upperBound() + eval::delta(player)*2;
      }
      else
      {
	// P ʤΤ Lower_bound  P ˤȤäƤ Upper_bound
	if (next_record && next_record->hasLowerBound(1))
	  value = next_record->lowerBound();
      }
      if (value != INVALID_VALUE)
	sorter.insert(std::make_pair
		      (value, MoveEntry(cur_move.getMove(), next_record)));
    }
  }
  showMoves();
}
  
osl::search::analyzer::
TableAnalyzer::TableAnalyzer(const SimpleHashTable& t, std::ostream& o) 
  : table(t), os(o)
{
}
osl::search::analyzer::
TableAnalyzer::~TableAnalyzer()
{
}
  
void osl::search::analyzer::
TableAnalyzer::examineTopLevel(const HashKey& key, int num_print_moves) const
{
  os << "analysis of root node\n";
  const SimpleHashRecord *record = table.find(key);
  if (! record)
  {
    os << "record not found\n";
    return;
  }

  if (record->lowerLimit() == SearchTable::CheckmateSpecialDepth)
  {
    if (! record->bestMove().getMove().isPass())
      os << "win by checkmate " << record->bestMove().getMove() << "\n";
    else
      os << "lose by checkmate\n";
    return;
  }

  TopLevelAnalyzer helper(key.turn(), &table, num_print_moves, os);
  helper.examineAll(record);
}
void osl::search::analyzer::
TableAnalyzer::examineUpperBounds(const HashKey& key, int num_print_moves) const
{
  os << "analysis of a node after the best move\n";
  const SimpleHashRecord *record = table.find(key);
  if (! record)
  {
    os << "record not found\n";
    return;
  }

  TopLevelAnalyzer helper(key.turn(), &table, num_print_moves, os);
  helper.examineAll(record);
}

void osl::search::analyzer::
TableAnalyzer::makeBestMoveList(const SimpleHashRecord *record, SearchMoveList& moves,
				int max)
{
  RecordSet dejavu;
  for (int i=0; i<max; ++i)
  {
    if (! record)
      break;
    const SearchMove best_move = record->bestMove();
    if ((! best_move.validMove())
	|| (dejavu.find(record) != dejavu.end()))
      break;
    moves.push_front(best_move); // ̵ͤξ MOVE_INVALID Ͽ
    if (best_move.getMove().isPass())
      break;
    dejavu.insert(record);
    record = best_move.record;
  }
  moves.reverse();
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
