/* historyTable.cc
 */
#include "osl/search/historyTable.h"
#include "osl/misc/carray3d.h"
#include "osl/move_generator/legalMoves.h"
#include "osl/container/moveVector.h"
#include <map>
#include <cmath>
#include <iostream>

const int osl::search::HistoryTable::maximum_logp;

struct osl::search::HistoryTable::Table
{
  /** player, from, to */
  typedef CArray3d<double,2,Position::SIZE,Position::SIZE> array_t;
  array_t moves;
  double total;
  Table() : total(0.0)
  {
    moves.fill(0.0);
  }
  void clear()
  {
    moves.fill(0.0);
    total = 0.0;
  }
  static int fromIndex(Move move)
  {
    const Position from = move.from();
    return from.isPieceStand() ? (int)move.ptype() : (int)from.index();
  }
  double& count(Move move)
  {
    return moves[playerToIndex(move.player())][fromIndex(move)]
      [move.to().index()];
  }
  const double& count(Move move) const 
  {
    return moves[playerToIndex(move.player())][fromIndex(move)]
      [move.to().index()];
  }
};

osl::search::
HistoryTable::HistoryTable()
  : table(new Table)
{
}

osl::search::
HistoryTable::~HistoryTable()
{
}

osl::search::HistoryTable& osl::search::
HistoryTable::operator=(const HistoryTable& o)
{
  if (this != &o)
    *table = *o.table;
  return *this;
}

void osl::search::
HistoryTable::clear()
{
  table->clear();
}

void osl::search::
HistoryTable::setMove(int depth_left, Move best_move)
{
  const double count = depth_left*depth_left;
  table->count(best_move) += count;
  table->total += count;
}

void osl::search::
HistoryTable::getMoves(const NumEffectState& state,
		       MoveLogProbVector& moves) const
{
  MoveVector plain_moves;
  LegalMoves::generate(state, plain_moves);

  typedef std::multimap<int, Move> sorter_t;
  sorter_t sorter;
  for (MoveVector::const_iterator p=plain_moves.begin(); 
       p!=plain_moves.end(); ++p)
  {
    const int logp = this->logp(*p);
    if (logp < maximum_logp)
      sorter.insert(std::make_pair(logp, *p));    
  }

  moves.clear();
  for (sorter_t::const_iterator p=sorter.begin(); 
       p!=sorter.end() && moves.size() < moves.capacity(); ++p)
  {
    moves.push_back(MoveLogProb(p->second, p->first));
  }
}

int osl::search::
HistoryTable::logp(double p)
{
  return static_cast<int>(-log(p)/log(2.0)*100);
}

int osl::search::
HistoryTable::logp(Move move) const
{
  const double& count = table->count(move);
  if (count < 1)
    return maximum_logp;
  return std::max(50,logp(sqrt(count) / sqrt(table->total)));
}

namespace osl
{
  namespace
  {
    const Move make_move(Player player, int i, int j)
    {
      const Position to = Position::makeDirect(j);
      Position from = Position::makeDirect(i);
      if (! from.isOnBoard())
      {
	const Ptype ptype = static_cast<Ptype>(i);
	assert(isBasic(ptype));
	return Move(to, ptype, player);
      }
      return Move(from, to, KING, PTYPE_EMPTY, false, player);
    }
  }
}

void osl::search::
HistoryTable::dump(std::ostream& os) const
{
  typedef CArray3d<double,2,Position::SIZE,Position::SIZE> array_t;
  typedef std::multimap<int, Move> sorter_t;
  sorter_t sorter;
  const double denominator = sqrt(table->total);
  for (int i=0; i<Position::SIZE; ++i)
  {
    for (int j=0; j<Position::SIZE; ++j)
    {
      double count = table->moves[playerToIndex(BLACK)][i][j];
      if (count >= 1)
      {
	const int logp = HistoryTable::logp(sqrt(count)/denominator);
	sorter.insert(std::make_pair(logp, make_move(BLACK, i, j)));
      }
      count = table->moves[playerToIndex(WHITE)][i][j];
      if (count >= 1)
      {
	const int logp = HistoryTable::logp(sqrt(count)/denominator);
	sorter.insert(std::make_pair(logp, make_move(WHITE, i, j)));
      }
    }
  }
  
  for (sorter_t::const_iterator p=sorter.begin(); p!=sorter.end(); ++p)
  {
    os << p->first << " " << p->second << "\n";
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
