/*
 * parallelSearch.cc
 */
#include "osl/search/parallelSearch.h"
#include <iostream>
#include <time.h>

#ifndef OSL_NCPUS
#  define OSL_NCPUS 4
#endif

osl::search::ParallelSearch osl::search::parallelSearch;

osl::search::StopNow::StopNow()
  : std::runtime_error("StopNow")
{
}

/* ------------------------------------------------------------------------- */

osl::search::DecActiveHelper::DecActiveHelper(ParallelSearch & parallel_search)
  : parallel_search(parallel_search){
  parallel_search.decActives();
}
osl::search::DecActiveHelper::~DecActiveHelper(){
  parallel_search.incActives();
}

osl::search::IncActiveHelper::IncActiveHelper(ParallelSearch & parallel_search)
  : parallel_search(parallel_search){
  parallel_search.incActives();
}
osl::search::IncActiveHelper::~IncActiveHelper(){
  parallel_search.decActives();
}

osl::search::IncWaitingHelper::IncWaitingHelper(ParallelSearch & parallel_search)
  : parallel_search(parallel_search){
  parallel_search.incWaiting();
}
osl::search::IncWaitingHelper::~IncWaitingHelper(){
  parallel_search.decWaiting();
}

osl::search::DecWaitingHelper::DecWaitingHelper(ParallelSearch & parallel_search)
  : parallel_search(parallel_search){
  parallel_search.decWaiting();
}
osl::search::DecWaitingHelper::~DecWaitingHelper(){
  parallel_search.incWaiting();
}

/* ------------------------------------------------------------------------- */

osl::search::JobContent::~JobContent()
{
}

void osl::search::JobContent::waitResult()
{
  boost::mutex::scoped_lock lk(mutex);
  if (isFinished())
    return;
  DecActiveHelper decActiveHelper(parallelSearch);
  parallelSearch.makeWorkers();
  condition.wait(lk);
  parallelSearch.checkStop();
}

void osl::search::JobContent::checkStop(){
  if (must_stop_now || parallelSearch.stopping()) {
    throw StopNow();
  }
}

void osl::search::JobContent::stopNow() {
  must_stop_now=true;
  if (boost::shared_ptr<JobContent> ptr = child.lock()) {
    if (! ptr->stopping())
      ptr->stopNow();
  }
}

/* ------------------------------------------------------------------------- */
void osl::search::Worker::operator()(void) {
  th=new boost::thread();
  for (;;) {
    try{
      Job job=parent->getJob(this);
      SetCurrentJobLock lock(*this,job);
      job();
    }
    catch (StopNow& e) {
      std::cerr << "ignore StopNow at root " << threadId() << "\n";
    }
    catch (std::exception& e) {
      std::cerr << "\ncaught exception at root " << threadId() << " "
		<< e.what() << "\n";
      stopMaster();
    }
    catch (...)
    {
      std::cerr << "\ncaught unknown exception at root " << threadId() << "\n";
      stopMaster();
    }
  }
}

void osl::search::Worker::stopMaster() {
  std::cerr << "\a";
  const bool success = parent->stopMaster();
  if (! success) {
    std::cerr << "stopMaster failed in worker " << threadId()
	      << std::endl;
  }
}

void osl::search::Worker::push_back(Job job) {
  {
    boost::mutex::scoped_lock lk(parent->shared_mutex);
    job_pool->push_back(job);
  }
  parent->makeWorkers();
}

bool osl::search::Worker::pop_back_job() {
  boost::mutex::scoped_lock lk(parent->shared_mutex);
  return job_pool->pop_back();
}

void osl::search::Worker::checkStop() const {
  if (!current_job.isNull()) { // sometimes, root worker has no job
    current_job.checkStop();
  }
  else if (parallelSearch.stopping()) {
    throw StopNow();
  }
}

/* ------------------------------------------------------------------------- */

osl::search::
ParallelSearch::ParallelSearch() 
  : num_actives(1), waiting(0), num_cpus(OSL_NCPUS), master_stop_flag(0),
    verbose(false)
{
#ifdef ONE_THREAD_PER_PROC
  workers.reserve(num_cpus);
#else
  workers.reserve(num_cpus*8);
#endif
  Worker *self=new Worker(this, new boost::thread());
  addInLock(self);
  assert(self == getWorker());
}

osl::search::
ParallelSearch::~ParallelSearch()
{
  if (verbose)
    std::cerr << "ParallelSearch"
	      << " max_num_actives " << max_num_actives.value()
	      << " workers " << workers.size() << std::endl;
}

const osl::search::Job osl::search::
ParallelSearch::getJobInLock(Worker *w)
{
  int min_priority=JobContent::NotFoundPriority;
  int min_i= -1;
  for (size_t i=0;i<workers.size();i++) {
    int priority=workers[i]->getPriority();
    if (priority<min_priority) {
      min_priority=priority;
      min_i=i;
    }
  }
  if (min_i>=0) {
    Job ret=workers[min_i]->getJob(w);
    assert(!ret.isNull());
    return ret;
  }
  return Job();
}

const osl::search::Job osl::search::
ParallelSearch::getJobNonBlock(Worker *w)
{
  w->checkStop();		// may throw
  Job ret;
  {
    boost::mutex::scoped_lock lk(shared_mutex);
    ret=getJobInLock(w);
  }
  return ret;
}

const osl::search::Job osl::search::
ParallelSearch::getJob(Worker *w)
{
  IncWaitingHelper incWaitingHelper(parallelSearch);
  DecActiveHelper decActiveHelper(parallelSearch);
  for (;;) {
    boost::mutex::scoped_lock lk(shared_mutex);
    Job ret=getJobInLock(w);
    if (!ret.isNull())
      return ret;
    shared_condition.wait(lk);
    w->checkStop();
  }
}

osl::search::Worker* osl::search::
ParallelSearch::getWorker() {
  boost::mutex::scoped_lock lk(shared_mutex);
  boost::thread self;
  for (size_t i=0;i<workers.size();i++) {
    boost::thread* th=workers[i]->getThread();
    if (th!=0 && *th==self) 
      return workers[i];
  }
  Worker *worker=new Worker(this,new boost::thread());
  addInLock(worker);
  return worker;
}

void osl::search::
ParallelSearch::checkStop() {
  Worker *worker=getWorker();
  assert(worker!=0);
  worker->checkStop();
}

void osl::search::
ParallelSearch::makeWorkers() {
  {
    boost::mutex::scoped_lock lk(shared_mutex);
#ifdef ONE_THREAD_PER_PROC
    while(thread_count.value()<num_cpus-1){
      thread_count.inc();
      incActives();
      Worker *worker=new Worker(this);
      addInLock(worker);
      new boost::thread(*worker);
    }
    if (waiting.value()!=0) 
      notify_one();
#else
    if (waiting.value()==0) {
      if (num_actives.value()<num_cpus) {
	thread_count.inc();
	incActives();
	Worker *worker=new Worker(this);
	addInLock(worker);
	new boost::thread(*worker);
      }
    }
    else if (num_actives.value()<num_cpus) {
      notify_one();
    }
#endif
  }
  checkTime();
}

void osl::search::
ParallelSearch::checkTime() {
  if (! verbose)
    return;
  time_t new_time=time(NULL);
  if (last_time != new_time) {
    last_time=new_time;
    if (num_actives.value() != num_cpus)
      std::cerr << "num_actives=" << num_actives.value() << 
	",waiting=" << waiting.value() << std::endl;
  }
}

void osl::search::
ParallelSearch::incActives() {
  num_actives.inc();
  max_num_actives.max(num_actives.value());
  checkTime();
}
void osl::search::
ParallelSearch::decActives() {
  checkTime();
  num_actives.dec();
}
void osl::search::
ParallelSearch::incWaiting() {
  waiting.inc();
}
void osl::search::
ParallelSearch::decWaiting() {
  waiting.dec();
}


void osl::search::
ParallelSearch::waitAll() {
  if (verbose)
    std::cerr << "ParallelSearch::waitAll\n";
  struct timespec t = { 0, 100000 };
  for (;;) {
    {
      boost::mutex::scoped_lock lk(shared_mutex);
      stop_all=true;

      for (size_t i=0;i<workers.size();i++) 
	workers[i]->clearJobs();

#ifdef ONE_THREAD_PER_PROC
      const int n=1+thread_count.value()-waiting.value();
#else
      const int n=num_actives.value();
#endif
      if (n<=1) 
	break;
    }
    nanosleep(&t, 0);
  }
  stop_all=false;
  master_stop_flag = AlarmSwitch(0);
}

bool osl::search::
ParallelSearch::stopMaster()
{
  if (! master_stop_flag.notify)
  {
    std::cerr << "ParallelSearch master_stop_flag is null\n";
    return false;
  }
  *master_stop_flag.notify = Alarm::TIMEOUT;
  return true;
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
