/* pawnMaskState.cc
 */
#include "osl/state/pawnMaskState.h"
#include "osl/state/pawnMaskState.tcc"
#include "osl/apply_move/applyMove.h"
#include "osl/apply_move/doUndoMoveLock.tcc"
#include "osl/record/csaIOError.h"
#include "osl/pieceTable.h"
#include <iostream>

namespace osl 
{
  // explicit instantiation
  template DoUndoMoveLock::DoUndoMoveLock(PawnMaskState&, Move);

} // namespace osl

osl::PawnMaskState::PawnMaskState(){
  init();
}

void osl::PawnMaskState::initPawnMask(){
  pawnMask[0].clearAll();
  pawnMask[1].clearAll();
  for(int num=PtypeTraits<PAWN>::indexMin;
      num< PtypeTraits<PAWN>::indexLimit; num++){
    Piece p=getPieceOf(num);
    Player player=p.owner();
    Position pos=p.position();
    if(!pos.isPieceStand() && !p.isPromotedNotKingGold()){
      if (isPawnMaskSet(player,pos.x()))
      {
	std::cerr << std::endl << *this << p << std::endl;
	throw CsaIOError("2FU!");
      }
      pawnMask[playerToIndex(player)].set(pos);
    }
  }
}

osl::PawnMaskState::PawnMaskState(osl::SimpleState const& st):SimpleState(st){
  initPawnMask();
}

osl::PawnMaskState::PawnMaskState(Handicap h){
  init(h);
}

void osl::PawnMaskState::init()
{
  SimpleState::init();
  pawnMask[0].clearAll();
  pawnMask[1].clearAll();
}
  

void osl::PawnMaskState::init(Handicap h){
  SimpleState::init(h);
  initPawnMask();
}
  

osl::PawnMaskState::~PawnMaskState(){}

void osl::PawnMaskState::setPiece(Player player,Position pos,Ptype ptype){
  int num;
  for (num=0;num<40;num++){
    if (!used_mask.test(num) && Piece_Table.getPtypeOf(num)==unpromote(ptype)
	&& (ptype!=KING || 
	    num==PtypeTraits<KING>::indexMin+playerToIndex(player))){
      used_mask.set(num);
	
      Piece p(player,ptype,num,pos);
      setPieceOf(num,p);
      setBoard(pos,p);
      if (pos==Position::STAND())
	standMask(player).set(num);
      else{
	if (ptype==PAWN)
	  pawnMask[playerToIndex(player)].set(pos);

      }
      return;
    }
  }
  std::cerr << "osl::PawnMaskState::setPiece! maybe too many pieces " 
	    << ptype << " " << pos << " " << player << "\n";
  abort();
}

void osl::PawnMaskState::setPieceAll(Player player){
  for (int num=0;num<40;num++){
    if (!used_mask.test(num)){
      used_mask.set(num);
      standMask(player).set(num);
      Piece p(player,Piece_Table.getPtypeOf(num),num,Position::STAND());
      setPieceOf(num,p);
    }
  }
}
// check

bool osl::PawnMaskState::isConsistent(bool showError) const
{
  if(!SimpleState::isConsistent(showError)) 
    return false;
  // pawnMask;
  {
    CArray<BitXmask,2> pawnMask1;
    pawnMask1[0].clearAll();
    pawnMask1[1].clearAll();
    for (int num=PtypeTraits<PAWN>::indexMin;
	 num<PtypeTraits<PAWN>::indexLimit;num++){
      if (isOnBoard(num)){
	Piece p=getPieceOf(num);
	if (!p.isPromotedNotKingGold()){
	  pawnMask1[playerToIndex(p.owner())].set(p.position());
	}
      }
    }
    if ((pawnMask[0]!=pawnMask1[0])
	|| (pawnMask[1]!=pawnMask1[1]))
    {
      if (showError) 
	std::cerr << "pawnMask "
		  << pawnMask[0] << "!=" << pawnMask1[0] 
		  << " || " <<  pawnMask[1] << "!=" << pawnMask1[1]
		  << std::endl;
      return false;
    }
  }
  return true;
}

bool osl::state::operator==(const PawnMaskState& st1,const PawnMaskState& st2)
{
  assert(st1.isConsistent(false));
  assert(st2.isConsistent(false));
  if (st1.pawnMask[0]!=st2.pawnMask[0]) return false;
  if (st1.pawnMask[1]!=st2.pawnMask[1]) return false;
  for (int y=1;y<=9;y++)
    for (int x=9;x>0;x--){
      Piece p1=st1.getPieceOnBoard(Position(x,y));
      Piece p2=st2.getPieceOnBoard(Position(x,y));
      if (p1.ptypeO()!=p2.ptypeO()) return false;
    }
  return true;
      
}

std::ostream& osl::state::operator<<(std::ostream& os,const PawnMaskState& state)
{
  os << static_cast<SimpleState>(state);
  return os;
}

void osl::PawnMaskState::doSimpleMove(Position from, Position to, int promoteMask)
{
  Piece oldPiece;
  int num;
  if(getTurn()==BLACK){
    apply_move::ApplyDoUndoSimpleMove<BLACK,PawnMaskState>::
      prologue(*this, from, to, promoteMask, oldPiece, num);
  }
  else{
    apply_move::ApplyDoUndoSimpleMove<WHITE,PawnMaskState>::
      prologue(*this, from, to, promoteMask, oldPiece, num);
  }

  if (promoteMask!=0 && num<PtypeTraits<PAWN>::indexLimit)
    clearPawn(getTurn(), from);
}

void osl::PawnMaskState::doDropMove(Position to,Ptype ptype)
{
  Piece oldPiece;
  int num, numIndex, numLow;
  if(getTurn()==BLACK){
    apply_move::ApplyDoUndoDropMove<BLACK,PawnMaskState>::
      prologue(*this, ptype, to, oldPiece, num, numIndex, numLow);
  }
  else{
    apply_move::ApplyDoUndoDropMove<WHITE,PawnMaskState>::
      prologue(*this, ptype, to, oldPiece, num, numIndex, numLow);
  }
  if (ptype==PAWN)
    setPawn(getTurn(),to);
}

void osl::PawnMaskState::doCaptureMove(Position from, Position to, Piece target,int promoteMask)
{
  Piece oldPiece;
  int num0, num1;
  int num1Index;
  mask_t num1Mask;
  Ptype capturePtype;
  if(getTurn()==BLACK){
    apply_move::ApplyDoUndoCaptureMove<BLACK,PawnMaskState>::
      prologue(*this, from, to, target, promoteMask,
	       capturePtype, oldPiece, num0, num1, num1Index, num1Mask);
  }
  else{
    apply_move::ApplyDoUndoCaptureMove<WHITE,PawnMaskState>::
      prologue(*this, from, to, target, promoteMask,
	       capturePtype, oldPiece, num0, num1, num1Index, num1Mask);
  }

  if (capturePtype==PAWN)
    clearPawn(alt(getTurn()),to);
  if (promoteMask!=0 && num0<PtypeTraits<PAWN>::indexLimit)
    clearPawn(getTurn(),from);
}

/* ------------------------------------------------------------------------- */
osl::apply_move::DoUndoMoveLockSimple<osl::PawnMaskState>::
DoUndoMoveLockSimple(PawnMaskState& s, 
		     Position f, Position t, int promoteMask)
  : state(s), from(f), to(t)
{
  if(s.getTurn()==BLACK){
    apply_move::ApplyDoUndoSimpleMove<BLACK,PawnMaskState>::
      prologue(state, from, to, promoteMask, oldPiece, num);
  }
  else{
    apply_move::ApplyDoUndoSimpleMove<WHITE,PawnMaskState>::
      prologue(state, from, to, promoteMask, oldPiece, num);
  }
  clearPawn = (promoteMask!=0) && (num < PtypeTraits<PAWN>::indexLimit);
  if (clearPawn)
    s.clearPawn(s.getTurn(),from);
}

osl::apply_move::DoUndoMoveLockSimple<osl::PawnMaskState>::
~DoUndoMoveLockSimple()
{
  if (clearPawn)
    state.setPawn(state.getTurn(),from);
  if(state.getTurn()==BLACK){
    apply_move::ApplyDoUndoSimpleMove<BLACK,PawnMaskState>::
      epilogue(state, from, to, oldPiece, num);
  }
  else{
    apply_move::ApplyDoUndoSimpleMove<WHITE,PawnMaskState>::
      epilogue(state, from, to, oldPiece, num);
  }
}

osl::apply_move::DoUndoMoveLockDrop<osl::PawnMaskState>::
DoUndoMoveLockDrop(PawnMaskState& s, Position t, Ptype ptype)
  : state(s), to(t)
{
  if(state.getTurn()==BLACK){
    apply_move::ApplyDoUndoDropMove<BLACK,PawnMaskState>::
      prologue(s, ptype, to, oldPiece, num, numIndex, numLow);
  }
  else{
    apply_move::ApplyDoUndoDropMove<WHITE,PawnMaskState>::
      prologue(s, ptype, to, oldPiece, num, numIndex, numLow);
  }
  pawnDrop = (ptype==PAWN);
  if (pawnDrop)
    s.setPawn(state.getTurn(),to);
}

osl::apply_move::DoUndoMoveLockDrop<osl::PawnMaskState>::
~DoUndoMoveLockDrop()
{
  if (pawnDrop)
    state.clearPawn(state.getTurn(),to);
  if(state.getTurn()==BLACK){
    apply_move::ApplyDoUndoDropMove<BLACK,PawnMaskState>::
      epilogue(state, to, oldPiece, num, numIndex, numLow);
  }
  else{
    apply_move::ApplyDoUndoDropMove<WHITE,PawnMaskState>::
      epilogue(state, to, oldPiece, num, numIndex, numLow);
  }
}

osl::apply_move::DoUndoMoveLockCapture<osl::PawnMaskState>::
DoUndoMoveLockCapture(PawnMaskState& s, 
		      Position f, Position t, Piece tg,
		      int promoteMask)
  : state(s), from(f), to(t), target(tg)
{
  if(state.getTurn()==BLACK){
    apply_move::ApplyDoUndoCaptureMove<BLACK,PawnMaskState>::
      prologue(s, from, to, target, promoteMask,
	       capturePtype, oldPiece, num0, num1, num1Index, num1Mask);
  }
  else{
    apply_move::ApplyDoUndoCaptureMove<WHITE,PawnMaskState>::
      prologue(s, from, to, target, promoteMask,
	       capturePtype, oldPiece, num0, num1, num1Index, num1Mask);
  }
  pawnCapture = (capturePtype==PAWN);
  if (pawnCapture)
    s.clearPawn(alt(s.getTurn()),to);
  pawnPromotion = (promoteMask!=0) && (num0<PtypeTraits<PAWN>::indexLimit);
  if (pawnPromotion)
    s.clearPawn(s.getTurn(),from);
}
osl::apply_move::DoUndoMoveLockCapture<osl::PawnMaskState>::
~DoUndoMoveLockCapture()
{
  if (pawnPromotion)
    state.setPawn(state.getTurn(), from);
  if (pawnCapture)
    state.setPawn(alt(state.getTurn()), to);
  if(state.getTurn()==BLACK){
    apply_move::ApplyDoUndoCaptureMove<BLACK,PawnMaskState>::
      epilogue(state, from, to, target, oldPiece, num0, num1, num1Index, num1Mask);
  }
  else{
    apply_move::ApplyDoUndoCaptureMove<WHITE,PawnMaskState>::
      epilogue(state, from, to, target, oldPiece, num0, num1, num1Index, num1Mask);
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
