#include "osl/category/standardMoveGenerator.h"
#include "osl/category/categoryListUtil.h"
#include "osl/category/categoryUtil.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/stat/histogram.h"
#include "osl/stat/average.h"
#include "osl/container/moveStack.h"
#include "osl/container/moveLogProbSet.h"
#include "osl/eval/progressEval.h"
#include "osl/state/numEffectState.h"
#include "osl/moveLogProb.h"
#include "osl/record/csaRecord.h"
#include "osl/record/csaIOError.h"
#include "osl/record/kanjiPrint.h"
#include "osl/apply_move/applyMove.h"
#include "osl/misc/perfmon.h"

#include <vector>
#include <map>
#include <string>
#include <boost/format.hpp>
#include <iostream>
#include <iomanip>
using namespace osl;
using namespace osl::category;

/**
 * @file 
 * ƥӤ¬
 */

void usage(const char *prog)
{
  using namespace std;
  cerr << "Usage: " << prog << " [-va] [-l limit] [-O order-limit] [-f skip] csafiles\n"
       << "-v verbose output\n"
       << "-V very verbose output\n"
       << "-a analyze all legal moves\n"
       << endl;
  exit(1);
}

int limit = 1200;
size_t first_skip = 0;
bool verbose = false;
bool very_verbose = false;
bool all_moves = false;

stat::Average moves, cycles, cycles_per_move, probs;
stat::Histogram moves_histogram(2, 10);

struct CategoryStat
{
  stat::Average cycles_per_move;
  stat::Average moves;
};
typedef std::map<const char *,CategoryStat> map_t;
map_t category_table;

void test_file(const char *filename);
void show_table(const map_t&);

int main(int argc, char **argv)
{
  const char *program_name = argv[0];
  bool error_flag = false;
  extern char *optarg;
  extern int optind;
    
  char c;
  while ((c = getopt(argc, argv, "l:f:Vvh")) != EOF)
  {
    switch(c)
    {
    case 'l':	limit = atoi(optarg);
      break;
    case 'f':	first_skip = atoi(optarg);
      break;
    case 'V':	very_verbose = true;
      // fall through
    case 'v':	verbose = true;
      break;
    default:	error_flag = true;
    }
  }
  argc -= optind;
  argv += optind;

  if (error_flag || (argc < 1))
    usage(program_name);

  for (int i=0; i<argc; ++i)
  {
    if (i % 128 == 0)
      std::cerr << '.';
    test_file(argv[i]);
  }

  std::cout << "\n"
	    << "average probability " << probs.getAverage() << "\n"
	    << "average moves/position " << moves.getAverage() << "\n";
  if (verbose) {
    moves_histogram.show(std::cout);
  }  
  std::cout << "average cycles/position " << cycles.getAverage() << "\n"
	    << "average cycles/position/move " << cycles_per_move.getAverage()
	    << "\n";
  show_table(category_table);
}

/* ------------------------------------------------------------------------- */

struct CategoryCount
{
  Move the_move;
  size_t& counter;
  int& prob;
  const char *& last_category;
  size_t& last_category_generated;
  template <bool is_king_escape>
  bool operator()(const char *category_name, MoveLogProbVector& moves) 
  {
    last_category = category_name;
    last_category_generated = moves.size();
    for (size_t i=0; i<moves.size(); ++i,++counter) {
      if (moves[i].getMove() == the_move) {
	prob = moves[i].getLogProb();
	return false;
      }
    }
    return true;
  }
};


size_t num_positions = 0;
void test_position(Move next_move, 
		   const NumEffectState& state, const MoveStack& history)
{
  eval::ProgressEval eval(state);
  SimpleHashRecord record;
  const category::CategoryEnv env(&state, limit, &history,
				  eval.progress32(), &record);
  size_t count=0;
  int prob = limit;
  const char *last_category = 0;
  size_t last_category_generated = 0;
  CategoryCount counter
    = { next_move, count, prob, last_category, last_category_generated };
  misc::PerfMon clock;
  CategoryListUtil::forEachCategory(StandardMoveGenerator(), env, counter);
  const size_t consumed = clock.stop();
  if (count) {
    moves.add(count);
    moves_histogram.add(count);
    cycles.add(consumed);
    cycles_per_move.add(consumed/count);
    probs.add(prob);
    assert(last_category);
    CategoryStat& stat = category_table[last_category];
    stat.cycles_per_move.add(consumed/count);
    stat.moves.add(last_category_generated);
    ++num_positions;
  }
}

void test_file(const char *filename)
{
  Record rec;
  try {
    rec = CsaFile(filename).getRecord();
  }
  catch (CsaIOError& e) {
    std::cerr << "skip " << filename <<"\n";
    std::cerr << e.what() << "\n";
    return;
  }
  catch (...) {
    throw;
  }
  
  NumEffectState state(rec.getInitialState());
  vector<osl::Move> moves=rec.getMoves();

  MoveStack history;

  for (size_t i=0; i<moves.size(); ++i) {
    const Move move = moves[i];
    assert(state.isValidMove(move));
    /** פäƤ */
    if (i >= first_skip) {
      test_position(moves[i], state, history);
    }
    ApplyMoveOfTurn::doMove(state, move);
    history.push(move);
  }
}

/* ------------------------------------------------------------------------- */

struct Order 
{
  const std::string name;
  template <class Head, class Tail>
  int find(CategoryList<Head,Tail>, int cur)
  {
    if (Head::getName() == this->name)
      return cur;
    return find(Tail(), cur+1);
  }
  template <class Head>
  int find (CategoryList<Head,NullCategory>, int cur)
  {
    assert(Head::getName() == this->name);
    return cur;
  }
  static int order(const char *name) 
  {
    Order work = { name };
    return work.find(StandardMoveGenerator(), 0);
  }
};

void show_table(const map_t& table)
{
  typedef std::map<int, std::pair<const char*, CategoryStat> > sorter_t;
  sorter_t sorter;
  for (map_t::const_iterator p=table.begin(); p != table.end(); ++p) {
    const size_t order = Order::order(p->first);
    sorter[order] = *p;
  }
  std::cout <<(boost::format("%22s  %12s %9s %9s\n")
	       % "category caused cut"
	       % "cycles/move"
	       % "position%" % "#moves");
  for (sorter_t::const_iterator p=sorter.begin(); p!=sorter.end(); ++p) {
    const CategoryStat& stat = p->second.second;
    std::cout << (boost::format("%22s  % 12.2f % 9.2f % 9.2f\n")
		  % p->second.first 
		  % stat.cycles_per_move.getAverage()
		  % (stat.cycles_per_move.numElements()*100.0/num_positions)
		  % stat.moves.getAverage());
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
