// historyToTable.t.cc
#include "osl/game_playing/historyToTable.h"
#include "osl/game_playing/gameState.h"
#include "osl/search/simpleHashTable.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/record/csaString.h"
#include "osl/record/record.h"
#include "osl/sennichite.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

using namespace osl;
using namespace osl::game_playing;

class HistoryToTableTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(HistoryToTableTest);
  CPPUNIT_TEST(testCurrentState);
  CPPUNIT_TEST(testDominance);
  CPPUNIT_TEST(testBestMoveInWin);
  CPPUNIT_TEST(testBestMoveInLose);
  CPPUNIT_TEST_SUITE_END();
public:
  void testCurrentState();
  void testDominance();
  void testBestMoveInWin();
  void testBestMoveInLose();
};

CPPUNIT_TEST_SUITE_REGISTRATION(HistoryToTableTest);

const int BLACK_WIN = 1000;
const int DRAW = 0;
const int WHITE_WIN = -1000;
const int limit = search::SearchTable::HistorySpecialDepth;

void HistoryToTableTest::testBestMoveInWin()
{
  const char *horizon_drop_string = 
    "P1-KY *  * -KI *  *  * -KE-KY\n"
    "P2 * -OU * -RY-KI *  * -KA * \n"
    "P3 * +TO * -FU * -FU * -FU * \n"
    "P4+HI * -GI-GI-FU *  *  * -FU\n"
    "P5+FU *  *  *  *  *  *  *  * \n"
    "P6 *  *  * +FU *  *  *  *  * \n"
    "P7 *  * -KA+GI+FU+FU+FU+FU+FU\n"
    "P8 *  *  *  * +KI+GI+KI+OU * \n"
    "P9+KY *  *  *  *  *  * +KE+KY\n"
    "P+00FU00FU\n"
    "P-00AL\n"
    "-\n"
    "-7483GI\n"
    "+0084FU\n"
    "-8374GI\n"
    "+8483TO\n";
  CsaString horizon_drop(horizon_drop_string);
  GameState state(horizon_drop.getInitialState());
  const vector<Move> moves = horizon_drop.getRecord().getMoves();
  for (size_t i=0; i<moves.size(); ++i)
  {
    CPPUNIT_ASSERT(state.state().isValidMove(moves[i]));
    CPPUNIT_ASSERT(state.pushMove(moves[i]).isNormal());
  }

  SimpleHashTable table;
  HistoryToTable::adjustTable(state, table, BLACK_WIN, DRAW, WHITE_WIN);
  
  const SimpleHashRecord *record = table.find(state.state().getHash());
  CPPUNIT_ASSERT(record);
  CPPUNIT_ASSERT(record->hasLowerBound(limit));
  CPPUNIT_ASSERT(record->hasUpperBound(limit));
  CPPUNIT_ASSERT_EQUAL(WHITE_WIN, record->lowerBound());
  CPPUNIT_ASSERT_EQUAL(WHITE_WIN, record->upperBound());
  CPPUNIT_ASSERT(record->bestMove().getMove().isValid());
  CPPUNIT_ASSERT_EQUAL(Move(Position(7,4),Position(8,3),
			    SILVER,PPAWN,false,WHITE),
		       record->bestMove().getMove());
}

void HistoryToTableTest::testBestMoveInLose()
{
  const char *horizon_drop_string = 
    "P1-KY *  * -KI *  *  * -KE-KY\n"
    "P2 *  *  * -RY-KI *  * -KA * \n"
    "P3 * -OU * -FU * -FU * -FU * \n"
    "P4+HI * -GI-GI-FU *  *  * -FU\n"
    "P5+FU *  *  *  *  *  *  *  * \n"
    "P6 *  *  * +FU *  *  *  *  * \n"
    "P7 *  * -KA+GI+FU+FU+FU+FU+FU\n"
    "P8 *  *  *  * +KI+GI+KI+OU * \n"
    "P9+KY *  *  *  *  *  * +KE+KY\n"
    "P+00FU00FU\n"
    "P-00AL\n"
    "+\n"
    "+0084FU\n"
    "-8382OU\n"
    "+8483TO\n"
    "-8283OU\n";
  CsaString horizon_drop(horizon_drop_string);
  GameState state(horizon_drop.getInitialState());
  const vector<Move> moves = horizon_drop.getRecord().getMoves();
  for (size_t i=0; i<moves.size(); ++i)
  {
    CPPUNIT_ASSERT(state.state().isValidMove(moves[i]));
    CPPUNIT_ASSERT(state.pushMove(moves[i]).isNormal());
  }

  SimpleHashTable table;
  HistoryToTable::adjustTable(state, table, BLACK_WIN, DRAW, WHITE_WIN);
  
  const SimpleHashRecord *record = table.find(state.state().getHash());
  CPPUNIT_ASSERT(record);
  CPPUNIT_ASSERT(record->hasLowerBound(limit));
  CPPUNIT_ASSERT(record->hasUpperBound(limit));
  CPPUNIT_ASSERT_EQUAL(WHITE_WIN, record->lowerBound());
  CPPUNIT_ASSERT_EQUAL(WHITE_WIN, record->upperBound());
  CPPUNIT_ASSERT(record->bestMove().getMove().isInvalid());
}

void HistoryToTableTest::testCurrentState()
{
  GameState state((SimpleState(HIRATE)));
  SimpleHashTable table;
  HistoryToTable::adjustTable(state, table, BLACK_WIN, DRAW, WHITE_WIN);

  const SimpleHashRecord *record = table.find(state.state().getHash());
  CPPUNIT_ASSERT_EQUAL((const SimpleHashRecord*)0, record);
}

void HistoryToTableTest::testDominance()
{
  GameState state(CsaString(
		    "P1-KY *  *  *  *  *  * +NY * \n"
		    "P2 * -OU-KI-KI *  *  *  * +RY\n"
		    "P3 * -GI-KE+KI *  *  *  * +HI\n"
		    "P4 *  * -FU-KY-FU *  * -FU * \n"
		    "P5-FU-FU * -KE * -FU *  *  * \n"
		    "P6 *  * +FU-FU+FU * -FU *  * \n"
		    "P7+FU+FU *  *  *  *  *  *  * \n"
		    "P8+KY+GI+GI-UM *  *  *  *  * \n"
		    "P9+OU+KE *  *  *  *  * +KE * \n"
		    "P-00FU00FU00FU00FU00FU00FU00GI00KA\n"
		    "P+00KI\n"
		    "+\n").getInitialState());
  SimpleHashTable table;
  HistoryToTable::adjustTable(state, table, BLACK_WIN, DRAW, WHITE_WIN);
  
  {
    // black have more pawn
    HashEffectState dominance_state(CsaString(
				      "P1-KY *  *  *  *  *  * +NY * \n"
				      "P2 * -OU-KI-KI *  *  *  * +RY\n"
				      "P3 * -GI-KE+KI *  *  *  * +HI\n"
				      "P4 *  * -FU-KY-FU *  * -FU * \n"
				      "P5-FU-FU * -KE * -FU *  *  * \n"
				      "P6 *  * +FU-FU+FU * -FU *  * \n"
				      "P7+FU+FU *  *  *  *  *  *  * \n"
				      "P8+KY+GI+GI-UM *  *  *  *  * \n"
				      "P9+OU+KE *  *  *  *  * +KE * \n"
				      "P-00FU00FU00FU00FU00FU00GI00KA\n"
				      "P+00FU00KI\n"
				      "+\n").getInitialState());
    const SimpleHashRecord *record = table.find(dominance_state.getHash());
    CPPUNIT_ASSERT(record);
    CPPUNIT_ASSERT(record->hasLowerBound(limit));
    CPPUNIT_ASSERT(record->hasUpperBound(limit));
    CPPUNIT_ASSERT_EQUAL(BLACK_WIN, record->lowerBound());
    CPPUNIT_ASSERT_EQUAL(BLACK_WIN, record->upperBound());
  }

  {
    // white have more gold
    HashEffectState dominance_state(CsaString(
				      "P1-KY *  *  *  *  *  * +NY * \n"
				      "P2 * -OU-KI-KI *  *  *  * +RY\n"
				      "P3 * -GI-KE+KI *  *  *  * +HI\n"
				      "P4 *  * -FU-KY-FU *  * -FU * \n"
				      "P5-FU-FU * -KE * -FU *  *  * \n"
				      "P6 *  * +FU-FU+FU * -FU *  * \n"
				      "P7+FU+FU *  *  *  *  *  *  * \n"
				      "P8+KY+GI+GI-UM *  *  *  *  * \n"
				      "P9+OU+KE *  *  *  *  * +KE * \n"
				      "P-00FU00FU00FU00FU00FU00FU00KI00GI00KA\n"
				      "+\n").getInitialState());
    const SimpleHashRecord *record = table.find(dominance_state.getHash());
    CPPUNIT_ASSERT(record);
    CPPUNIT_ASSERT(record->hasLowerBound(limit));
    CPPUNIT_ASSERT(record->hasUpperBound(limit));
    CPPUNIT_ASSERT_EQUAL(WHITE_WIN, record->lowerBound());
    CPPUNIT_ASSERT_EQUAL(WHITE_WIN, record->upperBound());
  }

}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
