#include "osl/piece.h"
#include "osl/pieceTable.h"
#include "osl/container/pieceMask.h"
#include "osl/misc/slowBsf.h"

#include <iomanip>
#include <iostream>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

class PieceTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(PieceTest);
  CPPUNIT_TEST(testShow);
  CPPUNIT_TEST(testSize);
  CPPUNIT_TEST(testBsf);
  CPPUNIT_TEST(testAddPosition);
  CPPUNIT_TEST(testIsEmpty);
  CPPUNIT_TEST(testIsEdge);
  CPPUNIT_TEST(testIsOnBoardBy);
  CPPUNIT_TEST(testPromote);
  CPPUNIT_TEST(testPieceMask);
  CPPUNIT_TEST_SUITE_END();
 public:
  void testSize();
  void testShow();
  void testBsf();
  void testAddPosition();
  void testIsEmpty();
  void testIsEdge();
  void testIsOnBoardBy();
  void testPromote();
  void testPieceMask();
};

using namespace osl;
using namespace osl::misc;
extern bool isShortTest;

CPPUNIT_TEST_SUITE_REGISTRATION(PieceTest);

void PieceTest::testSize(){
  CPPUNIT_ASSERT_EQUAL(sizeof(int), sizeof(Piece));
}

void PieceTest::testShow(){
  if (! isShortTest)
  {
    for(int num=0;num<Piece::SIZE;num++)
      std::cerr << num << ":" << Piece_Table.getPtypeOf(num) 
		<< std::setbase(10) << std::endl;
  }
}
void PieceTest::testBsf(){
  for(unsigned int i=1;i<i+11111;i+=11111){
    CPPUNIT_ASSERT(BitOp::bsf(i)==slowbsf(i));
  }
}
void PieceTest::testAddPosition(){
  Piece p1=Piece(BLACK,PPAWN,0,Position(7,5));
  CPPUNIT_ASSERT(p1.position()==Position(7,5));
  p1+=Offset(1,2);
  CPPUNIT_ASSERT(p1==Piece(BLACK,PPAWN,0,Position(8,7)));
  p1.setPosition(Position(6,4));
  CPPUNIT_ASSERT(p1==Piece(BLACK,PPAWN,0,Position(6,4)));
}
void PieceTest::testIsEmpty(){
  CPPUNIT_ASSERT(! Piece(BLACK,PPAWN,0,Position(8,7)).isEmpty());
  CPPUNIT_ASSERT(! Piece(WHITE,KING,30,Position(6,4)).isEmpty());
  CPPUNIT_ASSERT(Piece::EMPTY().isEmpty());
  CPPUNIT_ASSERT(! Piece::EDGE().isEmpty());
}
void PieceTest::testIsEdge(){
  CPPUNIT_ASSERT(! Piece(BLACK,PPAWN,0,Position(8,7)).isEdge());
  CPPUNIT_ASSERT(! Piece(WHITE,KING,30,Position(6,4)).isEdge());
  CPPUNIT_ASSERT(! Piece::EMPTY().isEdge());
  CPPUNIT_ASSERT(Piece::EDGE().isEdge());
}
void PieceTest::testIsOnBoardBy(){
  Piece p1=Piece(BLACK,PPAWN,0,Position(8,7));
  Piece p2=Piece(WHITE,KING,30,Position(6,4));
  Piece p3=Piece(BLACK,PAWN,0,Position::STAND());
  Piece p4=Piece(WHITE,ROOK,39,Position::STAND());
  CPPUNIT_ASSERT(p1.isOnBoardByOwner<BLACK>());
  CPPUNIT_ASSERT(!p1.isOnBoardByOwner<WHITE>());
  CPPUNIT_ASSERT(!p2.isOnBoardByOwner<BLACK>());
  CPPUNIT_ASSERT(p2.isOnBoardByOwner<WHITE>());
  CPPUNIT_ASSERT(!p3.isOnBoardByOwner<BLACK>());
  CPPUNIT_ASSERT(!p3.isOnBoardByOwner<WHITE>());
  CPPUNIT_ASSERT(!p4.isOnBoardByOwner<BLACK>());
  CPPUNIT_ASSERT(!p4.isOnBoardByOwner<WHITE>());
  CPPUNIT_ASSERT(!Piece::EMPTY().isOnBoardByOwner<BLACK>());
  CPPUNIT_ASSERT(!Piece::EMPTY().isOnBoardByOwner<WHITE>());
  CPPUNIT_ASSERT(!Piece::EDGE().isOnBoardByOwner<BLACK>());
  CPPUNIT_ASSERT(!Piece::EDGE().isOnBoardByOwner<WHITE>());
}
void PieceTest::testPromote(){
  Piece p1=Piece(BLACK,PAWN,0,Position(8,7));
  Piece p2=p1.promote();
  CPPUNIT_ASSERT(p2==Piece(BLACK,PPAWN,0,Position(8,7)));
  Piece p3=p2.unpromote();
  CPPUNIT_ASSERT(p1==p3);
  Piece p4=p1.checkPromote(0); 
  CPPUNIT_ASSERT(p1==p4);
  Piece p5=p1.checkPromote(1); 
  CPPUNIT_ASSERT(p2==p5);
  Piece p6=p2.checkUnpromote(0);
  CPPUNIT_ASSERT(p2==p6);
  Piece p7=p2.checkUnpromote(1);
  CPPUNIT_ASSERT(p1==p7);
}
void PieceTest::testPieceMask(){
  PieceMask pm;
  // std::cerr << pm << std::endl;
  for(int i=0;i<40;i++)
    pm.set(i);
  for(int i=5;i<30;i++)
    pm.reset(i);
#if OSL_WORDSIZE == 64
  CPPUNIT_ASSERT_EQUAL(mask_t::makeDirect(0xffc000001fuLL),pm.getMask(0));
#elif OSL_WORDSIZE == 32
  std::cerr << std::hex << mask_t::makeDirect(0xc000001f) << "\n";
  CPPUNIT_ASSERT_EQUAL(mask_t::makeDirect(0xc000001f),pm.getMask(0));
  CPPUNIT_ASSERT_EQUAL(mask_t::makeDirect(0xff),pm.getMask(1));
#endif
  for(int i=0;i<5;i++)
    CPPUNIT_ASSERT(pm.test(i));
  for(int i=5;i<30;i++)
    CPPUNIT_ASSERT(!pm.test(i));
  for(int i=30;i<40;i++)
    CPPUNIT_ASSERT(pm.test(i));
}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
